Require Import Lib.


(** * Syntax

      Here is the syntax for ReWire. *)

(** ** Monads and Types *)

(** Here is the syntax for layered state monads. *)

Inductive Eff : Type :=
  | EffNone : Eff
  | EffR    : Eff
  | EffW    : Eff
  | EffRW   : Eff.

Inductive SMo : Type :=
  | MIdentity : SMo  (* zero-layer state monad *)
  | MStateT   : Ty -> Eff -> SMo -> SMo

with Mo : Type :=
  | MReactT   : Ty -> Ty -> SMo -> Mo
  | MNonReact : SMo -> Mo

(** Here is the syntax for ``ordinary'' types. *)

with Ty : Type :=
  | TArrow   : Ty -> Ty -> Ty
  | TProd    : Ty -> Ty -> Ty
  | TSum     : Ty -> Ty -> Ty
  | TNil     : Ty
  | TMonadic : Mo -> Ty -> Ty.

(** Equality of Eff, Ty, Mo, SMo is deciable: *)

Lemma Eff_eq_dec : forall e e' : Eff, {e = e'} + {e <> e'}.
Proof.
  decide equality.
Qed.

Hint Resolve Eff_eq_dec.

Lemma  Ty_eq_dec : forall x y : Ty,  {x = y} + {x <> y} 
 with  Mo_eq_dec : forall m n : Mo,  {m = n} + {m <> n}
 with SMo_eq_dec : forall s r : SMo, {s = r} + {s <> r}.
Proof.
  decide equality. decide equality. decide equality.
Qed. 

(** Mutual induction schemes for types and monads will be useful. 
    **N.B. - These may be obsolete given that I have figured out how to 
             prove mutually dependent lemmas without the need for (new)
             induction. *)

Scheme Ty_Mo_SMo_ind := Induction for Ty  Sort Prop
  with Mo_Ty_SMo_ind := Induction for Mo  Sort Prop
  with SMo_Ty_Mo_ind := Induction for SMo Sort Prop.

Combined Scheme Ty_Mo_SMo_mutind from Ty_Mo_SMo_ind, Mo_Ty_SMo_ind, SMo_Ty_Mo_ind.

Definition on_io (P:Ty -> Prop) (M:Mo) : Prop :=
  match M with
  | MNonReact SM     => True
  | MReactT TI TO SM => P TI /\ P TO
  end.


Theorem Ty_Mo_SMo_mutind_better'
     : forall (P : Ty -> Prop) (P0 : Mo -> Prop) (P1 : SMo -> Prop),
       (forall T1 : Ty, P T1 -> forall T2 : Ty, P T2 -> P (TArrow T1 T2)) ->
       (forall T1 : Ty, P T1 -> forall T2 : Ty, P T2 -> P (TProd  T1 T2)) ->
       (forall T1 : Ty, P T1 -> forall T2 : Ty, P T2 -> P (TSum   T1 T2)) ->
       P TNil ->
       (forall M : Mo, (P0 M /\ on_io P M) ->
        forall T : Ty, P T ->
        P (TMonadic M T)) ->
       (forall TI : Ty,
        P TI ->
        forall TO : Ty, P TO -> forall SM : SMo, P1 SM -> P0 (MReactT TI TO SM)) ->
       (forall SM : SMo, P1 SM -> P0 (MNonReact SM)) ->
       P1 MIdentity ->
       (forall T : Ty,
        P T -> forall (E : Eff) (SM : SMo), P1 SM -> P1 (MStateT T E SM)) ->
       (forall T : Ty, P T) /\ (forall M : Mo, P0 M /\ on_io P M) /\ (forall SM : SMo, P1 SM).
Proof.
  intros;
  apply Ty_Mo_SMo_mutind; auto.
  intros.
  split.
  auto.
  constructor; auto.
  intros; split; auto.
  constructor; auto.
Qed.

Theorem Ty_Mo_SMo_mutind_better
     : forall (P : Ty -> Prop) (P0 : Mo -> Prop) (P1 : SMo -> Prop),
       (forall T1 : Ty, P T1 -> forall T2 : Ty, P T2 -> P (TArrow T1 T2)) ->
       (forall T1 : Ty, P T1 -> forall T2 : Ty, P T2 -> P (TProd  T1 T2)) ->
       (forall T1 : Ty, P T1 -> forall T2 : Ty, P T2 -> P (TSum   T1 T2)) ->
       P TNil ->
       (forall M : Mo, (P0 M /\ on_io P M) ->
        forall T : Ty, P T ->
        P (TMonadic M T)) ->
       (forall TI : Ty,
        P TI ->
        forall TO : Ty,  P TO  -> forall SM : SMo, P1 SM -> P0 (MReactT TI TO SM)) ->
       (forall SM : SMo, P1 SM -> P0 (MNonReact SM)) ->
       P1 MIdentity ->
       (forall T : Ty,
        P T -> forall (E : Eff) (SM : SMo), P1 SM -> P1 (MStateT T E SM)) ->
       (forall T : Ty, P T) /\ (forall M : Mo, P0 M) /\ (forall SM : SMo, P1 SM).
Proof.
  intros.
  pose proof (Ty_Mo_SMo_mutind_better' _ _ _ H H0 H1 H2 H3 H4 H5 H6 H7).
  destruct H8 as [? [? ?]].
  split; auto.
  split; auto.
  intro.
  pose proof (H9 M).
  destruct H11; auto.
Qed.

(** *** Relating state monads by permissiveness *)

Inductive Eff_lt : Eff -> Eff -> Prop :=
  | Eff_lt_None_None : Eff_lt EffNone EffNone
  | Eff_lt_None_R    : Eff_lt EffNone EffR
  | Eff_lt_None_W    : Eff_lt EffNone EffW
  | Eff_lt_None_RW   : Eff_lt EffNone EffRW
  | Eff_lt_R_R       : Eff_lt EffR EffR
  | Eff_lt_R_RW      : Eff_lt EffR EffRW
  | Eff_lt_W_W       : Eff_lt EffW EffW
  | Eff_lt_W_RW      : Eff_lt EffW EffRW
  | Eff_lt_RW_RW     : Eff_lt EffRW EffRW.

Hint Constructors Eff_lt.

Theorem Eff_lt_refl : forall E, Eff_lt E E.
Proof.
  intro E; destruct E; auto.
Qed.

Hint Resolve Eff_lt_refl.

Theorem Eff_lt_antisymm : forall E1 E2,
                            Eff_lt E1 E2 ->
                            Eff_lt E2 E1 ->
                            E1 = E2.
Proof.
  intros E1 E2 H1 H2; inversion H1; inversion H2; crush.
Qed.

Hint Resolve Eff_lt_antisymm.

Theorem Eff_lt_trans : forall E1 E2 E3,
                         Eff_lt E1 E2 ->
                         Eff_lt E2 E3 ->
                         Eff_lt E1 E3.
Proof.
  intros E1 E2 E3 H1 H2; inversion H1; inversion H2; crush.
Qed.

Hint Resolve Eff_lt_trans.

(** Partial order on state monads. M1 less permissive than M2 means that M1 has
    an identical store shape but fewer or the same permissions. *)

Inductive smo_less_permissive : SMo -> SMo -> Prop :=
  | LP_StateT    : forall T E1 E2 M1 M2,
                      Eff_lt E1 E2 -> 
                      smo_less_permissive M1 M2 ->
                      smo_less_permissive (MStateT T E1 M1)
                                          (MStateT T E2 M2)
  | LP_Identity  : smo_less_permissive MIdentity MIdentity.

Hint Constructors smo_less_permissive.

Theorem less_permissive_refl : forall M, smo_less_permissive M M.
Proof.
  intros; induction M; crush.
Qed.

Theorem less_permissive_antisymm : forall M1 M2, smo_less_permissive M1 M2 ->
                                                 smo_less_permissive M2 M1 ->
                                                 M1 = M2.
Proof.
  intros M1 M2 Hw12; induction Hw12; auto.
  intro H0; inversion H0; subst.
  pose proof (Eff_lt_antisymm _ _ H H3); subst.
  rewrite (IHHw12 H7); auto.
Qed.

Theorem less_permissive_trans: forall M1 M2 M3, smo_less_permissive M1 M2 ->
                                                smo_less_permissive M2 M3 ->
                                                smo_less_permissive M1 M3.
Proof.
  induction M1;
    intros M2 M3 H H0;
    inversion H; subst; inversion H0; subst; eauto.
Qed.

Inductive Eff_disjoint : Eff -> Eff -> Prop :=
  | Eff_disjoint_None_None : Eff_disjoint EffNone EffNone
  | Eff_disjoint_None_R    : Eff_disjoint EffNone EffR
  | Eff_disjoint_None_W    : Eff_disjoint EffNone EffW
  | Eff_disjoint_None_RW   : Eff_disjoint EffNone EffRW
  | Eff_disjoint_R_None    : Eff_disjoint EffR EffNone
  | Eff_disjoint_R_W       : Eff_disjoint EffR EffW
  | Eff_disjoint_W_None    : Eff_disjoint EffW EffNone
  | Eff_disjoint_W_R       : Eff_disjoint EffW EffR
  | Eff_disjoint_RW_None   : Eff_disjoint EffRW EffNone.

Hint Constructors Eff_disjoint.

Theorem Eff_disjoint_symm : forall E1 E2,
                              Eff_disjoint E1 E2 ->
                              Eff_disjoint E2 E1.
Proof.
  intros E1 E2 H; inversion H; auto.
Qed.

(* Relation on state monads. "M1 disjoint with M2" means that M1 and M2 have
   the same store shape and their permission sets are disjoint. *)
Inductive smo_disjoint : SMo -> SMo -> Prop :=
  | Disjoint_StateT    : forall T E1 E2 M1 M2,
                           smo_disjoint M1 M2 ->
                           Eff_disjoint E1 E2 ->
                           smo_disjoint (MStateT T E1 M1)
                                        (MStateT T E2 M2)
  | Disjoint_Identity  : smo_disjoint MIdentity MIdentity.

Hint Constructors smo_disjoint.

Theorem smo_disjoint_symm : forall M1 M2, smo_disjoint M1 M2 ->
                                          smo_disjoint M2 M1.
Proof.
  induction M1; intros ? ?H; inversion H; subst; auto.
  inversion H5; subst; auto.
Qed.

(** ** Terms and Configurations *)

Inductive tm : Type :=
  (* Lambda calculus *)
  | tvar     : id -> tm
  | tapp     : tm -> tm -> tm
  | tabs     : id -> Ty -> tm -> tm  (* ty is type of binder *)
  (* Data *)
  | tunit    : tm
  | tpair    : tm -> tm -> tm
  | tproj    : tm -> tm -> tm
  | tinl     : Ty -> tm -> tm
  | tinr     : Ty -> tm -> tm
  | tcase    : tm -> tm -> tm -> tm
  (* Monad operations (generic) *)
  | treturn  : Mo -> tm -> tm
  | tbind    : Ty -> tm -> tm -> tm
  | tlift    : Mo -> tm -> tm
  (* State monad operations *)
  | televate : SMo -> tm -> tm
  | tget     : SMo -> tm
  | tput     : SMo -> tm -> tm
  | trunst   : tm -> tm -> tm
  | trunid   : tm -> tm
  (* Reactive monad operations *)
  | tpause   : Mo -> Ty -> tm -> tm
  | tunfold  : Mo -> Ty -> Ty -> tm -> tm -> tm
  | trunre   : Ty -> tm -> tm.

(** The next lemma aims to address the naming conventions and subcases of
    the destruct tactic, to wit, they suck. *)

Infix "$" := tapp (at level 40).

Notation "f >>=[ T ] g"   := (tbind T f g) 
                              (at level 40, T at level 99, format "'[hv ' f  >>=[ T ]  g ']'").

Notation "\( x ':' T ) t" := (tabs x T t) 
                              (at level 40, x at level 99, format "'[hv ' \( x ':' T )  t ']'").

Lemma term_cases : forall P : tm -> Prop,
       (forall i : id, P (tvar i)) ->
       (forall t1 t2 : tm, P (tapp t1 t2)) ->
       (forall (i : id) (T : Ty) (t : tm), P (tabs i T t)) ->
       P tunit ->
       (forall t1 t2 : tm, 
                        P (tpair t1 t2)) ->
       (forall t1 t2 : tm,  
                        P (tproj t1 t2)) ->
       (forall (T : Ty) (t : tm), 
                        P (tinl T t)) ->
       (forall (T : Ty) (t : tm), 
                        P (tinr T t)) ->
       (forall t1 t2 t3 : tm,
                        P (tcase t1 t2 t3)) ->
       (forall (M : Mo) (t : tm), 
                        P (treturn M t)) ->
       (forall (T : Ty) (t1 t2 : tm),
                        P (tbind T t1 t2)) ->
       (forall (M : Mo) (t : tm),
                        P (tlift M t)) ->
       (forall (S : SMo) (t : tm), 
                        P (televate S t)) ->
       (forall S : SMo, 
                        P (tget S)) ->
       (forall (S : SMo) (t : tm),
                        P (tput S t)) ->
       (forall t1 t2 : tm, 
                        P (trunst t1 t2)) ->
       (forall t : tm, 
                        P (trunid t)) ->
       (forall (M : Mo) (T : Ty) (t : tm),
                        P (tpause M T t)) ->
       (forall (M : Mo) (TA TB : Ty) (t1 t2 : tm),
                        P (tunfold M TA TB t1 t2)) ->
       (forall (T : Ty) (t : tm), 
                        P (trunre T t)) ->
       forall (t : tm), P t.
Proof. intros; destruct t eqn:?;auto. Qed.

Ltac destruct_term t := pattern t; apply term_cases; clear t; intros.

  
Tactic Notation "term_cases" tactic(first) ident(c) :=
  first;
  [ Case_aux c "tvar"     | Case_aux c "tapp"     | Case_aux c "tabs"
  | Case_aux c "tunit"    | Case_aux c "tpair"    | Case_aux c "tproj"
  | Case_aux c "tinl"     | Case_aux c "tinr"     | Case_aux c "tcase" 
  | Case_aux c "treturn"  | Case_aux c "tbind"    | Case_aux c "tlift"   
  | Case_aux c "televate" | Case_aux c "tget"     | Case_aux c "tput" 
  | Case_aux c "trunst"   | Case_aux c "trunid"   | Case_aux c "tpause" 
  | Case_aux c "tunfold"  | Case_aux c "trunre"   ].

Definition store := list tm.
Definition configuration := (tm*store)%type.

(** * Values, as far as lambda-calculus reduction is concerned. *)

Inductive value : tm -> Prop :=
  | v_abs : forall x T t,
      value (tabs x T t)
  | v_unit : value tunit
  | v_pair : forall t1 t2,
      value t1 -> 
      value t2 -> 
      value (tpair t1 t2)
  | v_inl : forall T t,
      value t -> 
      value (tinl T t)
  | v_inr : forall T t,
      value t -> 
      value (tinr T t)
  | v_return : forall M t,
      value t ->
      value (treturn M t)
  | v_bind : forall T t1 t2,
      value t1 ->
      value t2 ->
      value (tbind T t1 t2)
  | v_lift : forall M t,
      value t ->
      value (tlift M t)
  | v_elevate : forall SM t,
      value t ->
      value (televate SM t)
  | v_get : forall SM,
      value (tget SM)
  | v_put : forall SM t,
      value t ->
      value (tput SM t)
  | v_runst : forall t1 t2,
      value t1 ->
      value t2 ->
      value (trunst t1 t2)
  | v_pause : forall M T t,
      value t ->
      value (tpause M T t)
  | v_unfold : forall M TA TB t1 t2,
      value t1 ->
      value t2 ->
      value (tunfold M TA TB t1 t2)
  | v_runre : forall T t1,
      value t1 ->
      value (trunre T t1).

Hint Constructors value.

Theorem value_dec : forall (t:tm), value t \/ ~ (value t).
  intro.
  induction t; try (right; intro H; inversion H; fail);
               try (left; constructor; fail);
               try (destruct IHt; [left; constructor; auto | right; intro Hval; inversion Hval; contradiction]; fail);
               try (destruct IHt1;
                    [ destruct IHt2; [ left; constructor; assumption
                                     | right; intro Hval; inversion Hval; contradiction ]
                    | right; intro Hval; inversion Hval; contradiction ];
                    fail).
Qed.

Ltac decide_value e := destruct (value_dec e) as [?Hval | ?Hnotval].

(** ** Done configurations (the monadic analog of values) *)

Inductive done_mo : configuration -> Prop :=
  | done_return   : forall M v Sto,   value v -> done_mo (treturn M v,Sto)
  | done_pause    : forall M T v Sto, value v -> done_mo (tpause M T v,Sto).

Hint Constructors done_mo.

Theorem done_mo_dec (co:configuration) : done_mo co \/ ~ (done_mo co).
  destruct co as [t Sto].
  induction t; try (right; intro H; inversion H; fail).
  Case "treturn".
    decide_value t.
    SCase "t is a value".
      left; constructor; assumption.
    SCase "t is not a value".
      right; intro Hdone; inversion Hdone; contradiction.
  Case "tpause".
    decide_value t0.
    SCase "t is a value".
      left; constructor; assumption.
    SCase "t is not a value".
      right; intro Hdone; inversion Hdone; contradiction.
Qed.


(** * Typing Judgments *)

(** ** For terms *)
Definition context := partial_map Ty.

Reserved Notation "Gamma '|--' t '\in' T" (at level 40).

Inductive has_type : context -> tm -> Ty -> Prop :=
  | T_Var  : forall Gamma x T,
                    Gamma x = Some T -> Gamma |-- tvar x \in T
  | T_Abs  : forall Gamma x T T' t,
             extend Gamma x T |-- t \in T' ->
                    Gamma |-- tabs x T t \in TArrow T T'
  | T_App  : forall T T' Gamma t1 t2,
               Gamma |-- t1 \in TArrow T T' ->
               Gamma |-- t2 \in T ->
               Gamma |-- tapp t1 t2 \in T'
  | T_Unit : forall Gamma,
                    Gamma |-- tunit \in TNil
  | T_Pair : forall Gamma T1 T2 t1 t2,
               Gamma |-- t1 \in T1 ->
               Gamma |-- t2 \in T2 ->
                 Gamma |-- tpair t1 t2 \in TProd T1 T2
  | T_Proj : forall Gamma T1 T2 T3 t t',
               Gamma |-- t \in (TProd T1 T2) ->
               Gamma |-- t' \in TArrow T1 (TArrow T2 T3) ->
                 Gamma |-- tproj t t' \in T3
  | T_Inl  : forall Gamma T1 T2 t,
               Gamma |-- t \in T1 ->
                 Gamma |-- tinl T2 t \in TSum T1 T2
  | T_Inr  : forall Gamma T1 T2 t,
               Gamma |-- t \in T2 ->
                 Gamma |-- tinr T1 t \in TSum T1 T2
  | T_Case : forall Gamma Tl Tr Tres ts tl tr,
               Gamma |-- ts \in TSum Tl Tr ->
               Gamma |-- tl \in TArrow Tl Tres ->
               Gamma |-- tr \in TArrow Tr Tres ->
                 Gamma |-- tcase ts tl tr \in Tres
  | T_Return  : forall Gamma T M t,
                  Gamma |-- t \in T ->
                    Gamma |-- treturn M t \in TMonadic M T
  | T_Bind    : forall Gamma t M T1 t' T2,
                  Gamma |-- t  \in TMonadic M T1 ->
                  Gamma |-- t' \in TArrow T1 (TMonadic M T2) ->
                    Gamma |-- tbind T2 t t' \in TMonadic M T2
  | T_LiftSt  : forall Gamma t SM E T T',
                  Gamma |-- t \in TMonadic (MNonReact SM) T' ->
                    Gamma |-- tlift (MNonReact (MStateT T E SM)) t \in TMonadic (MNonReact (MStateT T E SM)) T'
  | T_LiftRe  : forall Gamma t SM T TI TO,
                  Gamma |-- t \in TMonadic (MNonReact SM) T ->
                    Gamma |-- tlift (MReactT TI TO SM) t \in TMonadic (MReactT TI TO SM) T
  | T_Elevate : forall Gamma SM SM' t T,
                  smo_less_permissive SM SM' ->
                  Gamma |-- t \in TMonadic (MNonReact SM) T ->
                    Gamma |-- televate SM' t \in TMonadic (MNonReact SM') T
  | T_Get     : forall Gamma SM E T,
                  Eff_lt EffR E ->
                  Gamma |-- tget (MStateT T E SM) \in TMonadic (MNonReact (MStateT T E SM)) T
  | T_Put     : forall Gamma SM E t T,
                  Eff_lt EffW E ->
                  Gamma |-- t \in T ->
                    Gamma |-- tput (MStateT T E SM) t \in TMonadic (MNonReact (MStateT T E SM)) TNil
  | T_RunSt   : forall Gamma SM E t1 t2 TS T,
                  Gamma |-- t1 \in TMonadic (MNonReact (MStateT TS E SM)) T ->
                  Gamma |-- t2 \in TS ->
                    Gamma |-- trunst t1 t2 \in TMonadic (MNonReact SM) (TProd T TS)
  | T_RunId   : forall Gamma t T,
                  Gamma |-- t \in TMonadic (MNonReact MIdentity) T ->
                    Gamma |-- trunid t \in T
  | T_Pause   : forall Gamma SM TI TO T t,
                  Gamma |-- t \in TMonadic (MNonReact SM) (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))) ->
                    Gamma |-- tpause (MReactT TI TO SM) T t \in TMonadic (MReactT TI TO SM) T
  | T_Unfold  : forall Gamma SM TI TO TA TB t1 t2,
                  Gamma |-- t1 \in TB ->
                  Gamma |-- t2 \in TArrow TB (TMonadic (MNonReact SM) (TSum TA (TProd TO (TArrow TI TB)))) ->
                    Gamma |-- tunfold (MReactT TI TO SM) TA TB t1 t2 \in TMonadic (MReactT TI TO SM) TA
  | T_RunRe   : forall Gamma SM TI TO T t,
                  Gamma |-- t \in TMonadic (MReactT TI TO SM) T ->
                    Gamma |-- trunre T t \in TMonadic (MNonReact SM) (TSum T (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))))

where "Gamma '|--' t '\in' T" := (has_type Gamma t T).

Hint Extern 1 (has_type _ (@tapp _ _) _) => eapply T_App; auto. 
Hint Extern 1 (has_type _ (@tbind _ _ _) _) => eapply T_Bind; auto.

Lemma type_unique : forall t Gamma T1 T2, 
      Gamma |-- t \in T1 -> 
      Gamma |-- t \in T2 -> T1 = T2.
Proof.
  induction t; intros Gamma T1 T2 Ht1 Ht2;
  try (inversion Ht1; subst; inversion Ht2; subst; auto);
  try (rewrite -> H1 in H2; inv H2; fail);
  try (pose proof (IHt _ _ _ H3 H4); inv H; fail);
  try (pose proof (IHt2 _ _ _ H4 H6);
       pose proof (IHt1 _ _ _ H2 H3);
       inv H; inv H0; fail);
  try (pose proof (IHt _ _ _ H1 H2); inv H; fail).
  Case "tabs".
    pose proof (IHt _ _ _ H4 H5); inv H.
  Case "case".
    pose proof (IHt1 _ _ _ H3 H4);
    pose proof (IHt2 _ _ _ H5 H8);
    inv H; inv H0.
  Case "tbind".
    pose proof (IHt1 _ _ _ H4 H6);
    pose proof (IHt2 _ _ _ H5 H7);
    inv H; inv H0.
  Case "tlift (state)".
    pose proof (IHt _ _ _ H3 H6);
    inv H.
  Case "tlift (react)".
    pose proof (IHt _ _ _ H3 H6);
    inv H.
  Case "televate".
    pose proof (IHt _ _ _ H4 H6);
    inv H.
Qed.


(** ** For configurations *)
Inductive store_matches_mo : store -> Mo -> Prop :=
  | matches_mo_id     : store_matches_mo nil (MNonReact MIdentity)
  | matches_mo_statet : forall SM E T t Sto,
                          \empty |-- t \in T ->
                          store_matches_mo Sto (MNonReact SM) ->
                            store_matches_mo (t::Sto) (MNonReact (MStateT T E SM))
  | matches_mo_reactt : forall SM TI TO Sto,
                          store_matches_mo Sto (MNonReact SM) ->
                            store_matches_mo Sto (MReactT TI TO SM).

Hint Constructors store_matches_mo.

Reserved Notation "co '|>' T" (at level 40).

Inductive store_all_values : store -> Prop :=
  | sav_empty : store_all_values nil
  | sav_cons  : forall s Sto, value s -> store_all_values Sto -> store_all_values (s::Sto).

Inductive configuration_has_type : configuration -> Ty -> Prop :=
  | configuration_has_type_intro : forall t T M Sto,
                                     \empty |-- t \in TMonadic M T ->
                                     store_all_values Sto ->
                                     store_matches_mo Sto M ->
                                       (t,Sto) |> TMonadic M T
where "co |> T" := (configuration_has_type co T).

Hint Constructors configuration_has_type.

Lemma less_permissive_store_matches : forall SM1 SM2 Sto,
                                         smo_less_permissive SM1 SM2 ->
                                         store_matches_mo Sto (MNonReact SM1) ->
                                         store_matches_mo Sto (MNonReact SM2).
Proof.
  intros SM1 SM2 Sto.
  generalize dependent SM2. generalize dependent SM1.
  induction Sto; intros SM1 SM2 Hless Hmatches.
    Case "empty".
      inversion Hmatches; subst; inversion Hless; auto.
    Case "cons".
      inversion Hmatches; subst; inversion Hless; subst; constructor.
        assumption. apply (IHSto SM); auto.
Qed.

Lemma more_permissive_store_matches : forall SM1 SM2 Sto,
                                        smo_less_permissive SM2 SM1 ->
                                        store_matches_mo Sto (MNonReact SM1) ->
                                        store_matches_mo Sto (MNonReact SM2).
Proof.
  intros SM1 SM2 Sto.
  generalize dependent SM2. generalize dependent SM1.
  induction Sto; intros SM1 SM2 Hless Hmatches.
    Case "empty".
      inversion Hmatches; subst; inversion Hless; auto.
    Case "cons".
      inversion Hmatches; subst; inversion Hless; subst; constructor.
        assumption. apply (IHSto SM); auto.
Qed.

(*
Lemma typable_stronger :
  forall Gamma Gamma' T t,
    (forall x T, Gamma x = Some T -> Gamma' x = Some T) ->
    Gamma |-- t \in T ->
    Gamma' |-- t \in T.
Proof.
  intros Gamma Gamma' T t Hstronger HTt;
  generalize dependent Gamma'; generalize dependent Gamma; generalize dependent T.
  induction t; intros; try (inversion HTt; subst; econstructor; eauto; fail).
  Case "tabs".
    inv HTt.
    constructor.
    eapply IHt.
      eapply H4.
      intros.
      destruct (eq_id_dec i x).
        SCase "i = x".
          unfold extend.
          subst.
          unfold extend in H.
          rewrite eq_id in H.
          inversion H; subst.
          rewrite eq_id.
          reflexivity.
        SCase "i <> x".
          unfold extend in H.
          rewrite neq_id in H; auto.
          unfold extend.
          rewrite neq_id; auto.
Qed.*)

Theorem typable_empty_everywhere:
  forall Gamma T t,
    \empty |-- t \in T ->
    Gamma |-- t \in T.
Proof.
  assert (typable_stronger : forall Gamma Gamma' T t,
                               (forall x T, Gamma x = Some T -> Gamma' x = Some T) ->
                               Gamma |-- t \in T -> Gamma' |-- t \in T).
         intros Gamma Gamma' T t Hstronger HTt;
         generalize dependent Gamma'; generalize dependent Gamma; generalize dependent T.
         induction t; intros; try (inversion HTt; subst; econstructor; eauto; fail).
         Case "tabs".
              inv HTt.
              constructor.
              eapply IHt.
              eapply H4.
              intros.
              destruct (eq_id_dec i x).
              SCase "i = x".
                     unfold extend.
                     subst.
                     unfold extend in H.
                     rewrite eq_id in H.
                     inversion H; subst.
                     rewrite eq_id.
                     reflexivity.
              SCase "i <> x".
                    unfold extend in H.
                    rewrite neq_id in H; auto.
                    unfold extend.
                    rewrite neq_id; auto.  
  intros.
  assert ((forall (x : id) (T0 : Ty), \empty x = Some T0 -> Gamma x = Some T0)) by (intros; inversion H0).
  apply (typable_stronger \empty Gamma T t H0 H).
Qed.

Ltac smash' := smash2 idtac typable_empty_everywhere.

Ltac weakening :=
  match goal with
    | [H: (\empty |-- ?t \in ?T)  |- (?Ctxt |-- ?t \in ?T) ] => apply (typable_empty_everywhere Ctxt T t H)
    | [ |- ?X ] => idtac X
  end.

Ltac type_unicity := 
  let rec goty :=
      let HeqTy := fresh "HeqTy" in (
      match goal with 
      | [H1: (?Ctxt |-- ?t \in ?T1), H2: (?Ctxt |-- ?t \in ?T2) |- _ ] => 
             pose proof (type_unique t Ctxt T1 T2 H1 H2) as HeqTy; hide hyp H1; goty
      | _ => show all hyps ; idtac "Done matching terms with syntactically distinct typing judgments in the current context."
      end)
   in goty.

(** Some lemmas about canonical forms. *)
Lemma canonical_forms_fun : forall t T1 T2,
      empty |-- t \in (TArrow T1 T2) ->
      value t ->
      exists x u, t = tabs x T1 u.
Proof.
  intros ? ? ? HT HVal. 
  inv HVal; inv HT; eexists; eauto.
Qed.

Lemma canonical_forms_pair : forall t T1 T2,
      empty |-- t \in (TProd T1 T2) ->
      value t ->
      exists t1 t2, t = tpair t1 t2.
Proof.
  intros ? ? ? HT HVal.
  inv HVal; inv HT; eexists; eauto.
Qed.

Lemma canonical_forms_copair : forall t T1 T2,
      empty |-- t \in (TSum T1 T2) ->
      value t ->
      (exists t', t = tinl T2 t') \/ (exists t', t = tinr T1 t').
Proof.
  intros ? ? ? HT HVal.
  inv HVal; inv HT; [ left;eexists;eauto | right;eexists;eauto ].
Qed.

Lemma canonical_forms_tunit : forall t, 
      empty |-- t \in TNil -> 
      value t ->
      t = tunit.
Proof. induction t; intros Hval HTv; inv Hval;inv HTv.
Qed.
