Require Import Lib.
Require Import ReWire_Base.
Require Import ReWire_Subst.
Require Import ReWire_Step.
Require Import ReWire_Progress.
Require Import ReWire_Preservation.
      
(** Lambda calculus normal forms reduce no further. *)

Definition normal_form (t:tm) : Prop := 
           ~ exists t', t ~~> t'.

(** Because lambda calculus values do not single step reduce, 
    it follows that if t is a value, then t is a normal form. *)

Lemma value_norm_form : forall (t : tm),
      value t -> normal_form t.
Proof.
    intros t Hval; induction Hval; intros [t' Hval']; inversion Hval'; eauto.
Qed.

(** For any term t, [t] halts iff there exists a value t', such that t ~~>* t'. *)

Definition halts (t:tm) : Prop :=  
           exists t', t ~~>* t' /\ value t'.

(** For any configuration c, [c] halts iff there exists a done configuration c', 
    such that c ~~~>* c'.  *)

Definition halts_mo (co:configuration) : Prop :=
           exists co', co ~~~>* co' /\ done_mo co'.

Lemma values_halt : forall t, 
      value t -> halts t.
Proof.
   intros t H. unfold halts.
   exists t. split; eauto.
Qed.

Lemma done_halts : forall co, 
      done_mo co -> halts_mo co.
Proof.
   intros. unfold halts_mo.
   exists co. split; eauto.
Qed.

(*** Our approach, following SF, to establishing strong (lambda and monadic) normalization, uses 
    the logical relations proof technique. The collection [R], of predicates indexed by type 
    expressions,  *)

CoInductive along_react : (store -> Prop) -> (tm -> Prop) -> (tm -> Prop) -> (tm -> Prop) -> configuration -> Prop :=
  | along_return  : forall (PS:store -> Prop) (PI PO PR:tm -> Prop) TI TO SM t Sto t' Sto',
                      (t,Sto) ~~~>* (treturn (MReactT TI TO SM) t',Sto') ->
                      value t' ->
                      PR t'    ->
                      PS Sto'  ->
                      along_react PS PI PO PR (t,Sto)
  | along_pause  : forall (PS:store -> Prop) (PI PO PR:tm -> Prop) TI TO SM T t Sto t' Sto' vl vr Sto'',
                      (t,Sto) ~~~>* (tpause (MReactT TI TO SM) T t',Sto') ->
                      value t' ->
                      value vl ->
                      value vr ->
                      PS Sto'  ->
                      PS Sto'' ->
                      (t',Sto') ~~~>* (treturn (MNonReact SM) (tpair vl vr),Sto'') ->
                      PO vl ->
                      (forall t'', PI t'' -> halts (tapp vr t'')) ->
                      (forall t'' Sto'',
                         PI t'' ->
                         PS Sto'' ->
                         along_react PS PI PO PR ((tapp vr t''),Sto'')) ->
                      along_react PS PI PO PR (t,Sto).

(** Our approach, following SF, to establishing strong (lambda and monadic) normalization, uses 
    the logical relations proof technique. A logical relation R 3.r.t. some property P The collection [R T], of predicates indexed by type 
    expressions,  *)

Fixpoint R (T:Ty) (t:tm) {struct T} : Prop :=
   \empty |-- t \in T /\ halts t /\
   match T with 
   | TArrow T1 T2                   => forall s, R T1 s -> R T2 (tapp t s)
   | TProd T1 T2                    => exists t1 t2,
                                         t ~~>* (tpair t1 t2) /\
                                         value t1 /\
                                         value t2 /\
                                         R T1 t1 /\
                                         R T2 t2
   | TSum T1 T2                     => exists t',
                                         value t' /\
                                         ((t ~~>* tinl T2 t' /\ R T1 t') \/
                                          (t ~~>* tinr T1 t' /\ R T2 t'))
   | TNil                           => True
   | TMonadic (MNonReact SM) T'     => forall Sto,
                                         Rsto SM Sto ->
                                         exists t' Sto',
                                           (t,Sto) ~~~>* (treturn (MNonReact SM) t',Sto') /\
                                           value t' /\
                                           R T' t' /\
                                           Rsto SM Sto'
   | TMonadic (MReactT TI TO SM) T' => forall Sto, 
                                         Rsto SM Sto ->
                                         exists t' Sto',
                                           (t,Sto) ~~~>* (t',Sto') /\
                                           value t' /\
                                           Rsto SM Sto' /\
                                            along_react (Rsto SM) (R TI) (R TO) (R T') (t',Sto')
     end
with Rsto (SM:SMo) (Sto:store) {struct SM} : Prop :=
   store_all_values Sto /\
   store_matches_mo Sto (MNonReact SM) /\
   match SM with
   | MIdentity       => True
   | MStateT T b SM' => exists t Sto',
                          R T t /\
                          Rsto SM' Sto' /\
                          Sto = t::Sto'
   end.

(** For any lambda calculus term t of type T, membership of t in [R T] entails that t halts. 
    Similarly, for any configuration c = (t,s), and monadic type M T, membership of t in 
    [R M T], and s in [Rsto] entails that c halts.   *) 

Lemma Rsto_all_values : forall {SM} {Sto},
      Rsto SM Sto -> store_all_values Sto.
Proof. 
   intros. destruct SM; inv H. 
Qed.

Lemma R_halts : forall {T} {t}, 
      R T t -> halts t. 
Proof.
  intros. destruct T; unfold R in H; inversion H; inversion H1;  assumption.
Qed.

Lemma R_halts_nonreact : forall T SM t Sto, 
      R (TMonadic (MNonReact SM) T) t ->
             Rsto SM Sto ->
          halts_mo (t, Sto).
Proof.
 intros. destruct H as [? [? ?]]. destruct (H2 _ H0) as [t' [Sto' [? [? [? ?]]]]]. 
         unfold halts_mo. 
         exists (treturn (MNonReact SM) t', Sto'). 
                 split; eauto.
Qed.

Lemma R_halts_react : forall TI TO SM T t Sto, 
      R (TMonadic (MReactT TI TO SM) T) t ->
      Rsto SM Sto ->
      halts_mo (t,Sto).
Proof.
  intros. destruct H as [? [? ?]]. destruct (H2 _ H0) as [t' [Sto' [? [? [? ?]]]]].
          inversion H6;subst.
          unfold halts_mo.
          exists (treturn (MReactT TI0 TO0 SM0) t'0, Sto'0).
                 split; eauto.
          exists (tpause (MReactT TI0 TO0 SM0) T0 t'0, Sto'0).
                 split; eauto.
Qed.

Lemma R_typable_empty : forall {T} {t}, 
               R T t -> 
      \empty |-- t \in T.
Proof. 
  intros. 
  destruct T; unfold R in H; inversion H; inversion H1; assumption. 
Qed.


(* ###################################################################### *)
(** *** Facts conerning Rsto and the permissiveness ordering on state monads. *)

Lemma Rsto_matches : forall Sto SM,
      Rsto SM Sto ->
      store_matches_mo Sto (MNonReact SM).
Proof.
  intros Sto SM; destruct SM; intro H; destruct H as [? [? ?]]; auto.
Qed.

Lemma Rsto_less_permissive : forall SM SM' Sto, 
      smo_less_permissive SM SM' -> 
      Rsto SM Sto ->
      Rsto SM' Sto.
Proof.
  induction SM.
    intros. inversion H; subst. simpl. pose proof (Rsto_matches _ _ H0). auto.
    intros SM' Sto Hlp Hrsto.
      simpl in Hrsto.
      destruct Hrsto as [Hav [Hmatch [t0 [Sto' [Hlcrt0 [HrstoSto' Heq]]]]]].
      subst.
      destruct SM'.
        inversion Hlp.
        simpl.
        split. auto.
        split. eapply less_permissive_store_matches. eauto. assumption.
        assert (t1=t). inversion Hlp; auto.
        subst.
        exists t0, Sto'.
        repeat (split; auto).
        eapply IHSM.
        inversion Hlp; auto.
        assumption.
Qed.

Lemma Rsto_more_permissive : forall SM SM' Sto, 
      smo_less_permissive SM' SM -> 
      Rsto SM Sto -> 
      Rsto SM' Sto.
Proof.
  induction SM.
    intros. inversion H; subst. simpl. pose proof (Rsto_matches _ _ H0). auto.
    intros SM' Sto Hlp Hrsto.
      simpl in Hrsto.
      destruct Hrsto as [Hav [Hmatch [t0 [Sto' [Hlcrt0 [HrstoSto' Heq]]]]]].
      subst.
      destruct SM'.
        inversion Hlp.
        simpl.
        split. auto.
        split. eapply more_permissive_store_matches. eauto. assumption.
        assert (t1=t). inversion Hlp; auto.
        subst.
        exists t0, Sto'.
        repeat (split; auto).
        eapply IHSM.
        inversion Hlp; auto.
        assumption.
Qed.


(* ###################################################################### *)
(** *** Lemmas concerning how the step and multistep relation interact 
         with halting and membership in R and the coinductive predicate [along2_react].  *)

Lemma step_preserves_halting : forall t t', 
      (t ~~> t') -> (halts t <-> halts t'). 
Proof.
intros t t' ST.  unfold halts. 
 split. 
 Case "->". 
  intros [t'' [STM V]]. 
  inversion STM; subst. 
   apply ex_falso_quodlibet.  apply value_norm_form in V. unfold normal_form in V. apply V. exists t'. auto. 
   rewrite (step_deterministic _ _ _ ST H). exists t''. split; assumption.
 Case "<-".
  intros [t'0 [STM V]]. 
  exists t'0. split; eauto. 
Qed.

Lemma multistep_preserves_halting : forall t t',
      (t ~~>* t') -> (halts t <-> halts t').
Proof.
  intros t t' Hstep.
  induction Hstep.
  Case "0-step".
    split; auto.
  Case "n-step".
    split.
    SCase "->".
      intro Hhx.
      unfold halts in Hhx.
      destruct Hhx as [t' [Hstepx Hvalt']].
      destruct Hstepx.
        contradict H; eapply value_not_step; eauto.
        pose proof (step_deterministic _ _ _ H H0); subst.
        apply IHHstep.
        unfold halts.
        exists z0; auto.
    SCase "<-".
      intro Hhz.
      destruct IHHstep.
      pose proof (H1 Hhz).
      destruct H2 as [t [Hstepy Hvalt]].
      exists t.
      split; auto.
      eapply rtc_step; eauto.
Qed.

Lemma step_preserves_along :
  forall t Sto t' Sto' (PS:store -> Prop) PI PO PR,
    (t,Sto) ~~~> (t',Sto') ->
    along_react PS PI PO PR (t',Sto') ->
    along_react PS PI PO PR (t,  Sto).
Proof.
  intros ? ? ? ? ? ? ? ? Hstep Halong.
  inversion Halong; subst.
  Case "to return".
    eapply along_return; eauto.
  Case "to pause". 
    eapply along_pause. eauto. auto. eapply H3. eapply H4. auto. eapply H6. auto. auto. auto.
    auto.
Defined.

Lemma step_preserves_along' :
  forall t Sto t' Sto' TI TO TP M,
    (t,Sto) ~~~> (t',Sto') ->
    along_react (Rsto M) (R TI) (R TO) (R TP) (t,Sto) ->
    along_react (Rsto M) (R TI) (R TO) (R TP) (t',Sto').
Proof.
  intros ? ? ? ? ? ? ? ? Hstep Halong.
  inversion Halong; subst.
  Case "to return".
    eapply along_return; eauto.
    assert (done_mo (treturn (MReactT TI0 TO0 SM) t'0, Sto'0)) by (constructor;auto).
    pose proof (step_done_second (t,Sto) (t',Sto') _ H Hstep H1).
    eapply H0.
  Case "to pause".
    assert (done_mo (tpause (MReactT TI0 TO0 SM) T t'0, Sto'0)) by (constructor;auto).
    pose proof (step_done_second (t,Sto) (t',Sto') _ H Hstep H1).
    eapply along_pause; try eapply H0. 
    auto. eapply H3. eapply H4. auto. eapply H6. auto. auto. auto. auto.
Defined.


Lemma step_preserves_along_star :
  forall t Sto t' Sto' (PS:store -> Prop) PI PO PR,
    (t,Sto) ~~~>* (t',Sto') ->
    along_react PS PI PO PR (t',Sto') ->
    along_react PS PI PO PR (t,  Sto).
Proof.
  intros ? ? ? ? ? ? ? ? Hstep.
  remember (t,Sto) as co.
  remember (t',Sto') as co'.
  generalize dependent Sto' ;   generalize dependent t';   generalize dependent Sto;   generalize dependent t.
  induction Hstep; intros; subst.
  Case "0-step".
    auto.
  Case "n-step".
    destruct y as (t1,Sto1).
    pose proof (IHHstep _ _ eq_refl _ _ eq_refl H0).
    eapply step_preserves_along; eauto.
Defined.


Lemma step_preserves_along_star' :
  forall t Sto t' Sto' TI TO TP M,
    (t,Sto) ~~~>* (t',Sto') ->
    along_react (Rsto M) (R TI) (R TO) (R TP) (t,  Sto) ->
    along_react (Rsto M) (R TI) (R TO) (R TP) (t',Sto').
Proof.
  intros ? ? ? ? ? ? ? ? Hstep Halong;
  remember (t,Sto) as co;
  remember (t',Sto') as co';
  generalize dependent Sto;   generalize dependent t;   generalize dependent Sto';   generalize dependent t'.
  induction Hstep; intros; subst.
  Case "0-step".
    auto.
  Case "n > 0 step".
    destruct y as (t1, Sto1).
    pose proof (step_preserves_along' _ _ _ _ _ _ _ _ H Halong).
    pose proof (IHHstep H0 _ _ eq_refl _ _ eq_refl).
    assumption.
Defined.


Lemma step_along_react :
  forall t Sto t' Sto' TI TO TP M,
    (t,Sto) ~~~> (t',Sto') ->
    (along_react (Rsto M) (R TI) (R TO) (R TP) (t, Sto) <->
     along_react (Rsto M) (R TI) (R TO) (R TP) (t',Sto')).
Proof.
  intros ? ? ? ? ? ? ? ? Hstep.
  split; intros.
  Case "->".
    eapply step_preserves_along'; eauto.
  Case "<-". 
    eapply step_preserves_along; eauto.
Defined.

Lemma step_along_react_star :
  forall t Sto t' Sto' TI TO TP M,
    (t,Sto) ~~~>* (t',Sto') ->
    (along_react (Rsto M) (R TI) (R TO) (R TP) (t, Sto) <->
     along_react (Rsto M) (R TI) (R TO) (R TP) (t',Sto')).
Proof.
  intros ? ? ? ? ? ? ? ? Hstep.
  split; intros.
  Case "->".
    eapply step_preserves_along_star'; eauto.
  Case "<-". 
    eapply step_preserves_along_star; eauto.
Defined.

Lemma step_preserves_R : forall T t t', 
      (t ~~> t') -> 
         R T t -> 
         R T t'.
Proof.
  induction T; intros t t' tSt' Rt; unfold R; fold R; unfold R in Rt; fold R in Rt;
  destruct Rt as [tt [halts_t RRt]].
  Case "TArrow".
   split. eapply preservation; eauto.   
   split. apply (step_preserves_halting _ _ tSt'); eauto. 
   intros. 
   eapply IHT2. 
   apply  ST_App1. apply tSt'. 
   apply RRt; auto.
  Case "TProd".
    split. eapply preservation; eauto.
    split. apply (step_preserves_halting _ _ tSt'); eauto.
    destruct RRt as [t1 [t2 [STM [v_t1 [v_t2 extra]]]]]. 
    exists t1. 
    exists t2.
    split. inversion STM; subst.
    exfalso. 
       assert (value (tpair t1 t2)) as valpair. 
              constructor; auto.
       apply (value_norm_form _ valpair). 
    exists t'; auto.
    rewrite (step_deterministic _ _ _ tSt' H). assumption. auto.
  Case "TSum".
    split. eapply preservation; eauto.
    split. apply (step_preserves_halting _ _ tSt'); eauto.
    destruct RRt as [t1 [v_t1 STM]].
    exists t1.
    destruct STM as [H | H]; split; eauto; [left | right]; inversion H; split; try assumption.
    SCase "going to inl".
      inversion H0; subst.
      SSCase "zero-step".
        assert (value (tinl T2 t1)).
          constructor.
          assumption.
        contradict tSt'.
          apply value_not_step; assumption.
      SSCase "more then zero-step".
        rewrite (step_deterministic _ _ _ tSt' H2). assumption.
    SCase "going to inr".
      inversion H0; subst.
      SSCase "zero-step".
        assert (value (tinr T1 t1)).
          constructor.
          assumption.
        contradict tSt'.
          apply value_not_step; assumption.
      SSCase "more than zero-step".
        rewrite (step_deterministic _ _ _ tSt' H2). assumption.
  Case "TNil". 
    split. eapply preservation; eauto.
    split. apply (step_preserves_halting _ _ tSt'); eauto.
    assumption.
  Case "TMonadic".
    fold Rsto. fold Rsto in RRt.
    split. preservation. auto. 
    split. apply (step_preserves_halting _ _ tSt'); eauto.
    destruct m.
    SCase "reactive".
      intros.
      pose proof (RRt Sto H).
      destruct H0 as [t'0 [Sto' ?]]. 
      exists t'0. exists Sto'.
      destruct H0 as [? [? [? ?]]].
      split; auto.
      eapply step_monad_second; eauto.
    SCase "nonreactive".
      intros.
      assert ((t,Sto) ~~~>* (t',Sto)) by (eapply STM_LC in tSt'; eapply rtc_R_mo in tSt';eauto).
      pose proof (RRt Sto H). 
      destruct H1 as [t'0 [Sto' [? [? [? ?]]]]].
      exists t'0; exists Sto'. 
      repeat split; [eapply step_same_done; [constructor 
                                            |eapply H0 
                                            |assumption] 
                    |assumption
                    |assumption | assumption ]. 
                     assumption.
Qed.

Lemma multistep_preserves_R : forall T t t', 
      (t ~~>* t') -> 
          R T t -> 
          R T t'.
Proof. 
  intros T t t' STM; induction STM; intros. 
  assumption.
  apply IHSTM. eapply step_preserves_R. apply H. assumption.
Qed.

Lemma step_preserves_R' : forall T t t', 
      \empty |-- t \in T -> 
                t ~~> t' ->
                R T t'   -> 
                R T t. 
Proof.
  intros T. induction T; intros t t' Ht tSt' Rt'; unfold R; fold R; unfold R in Rt'; fold R in Rt';
  destruct Rt' as [_ [halts_t' RRt']]; split; try apply Ht.
Case "TArrow".
    split. apply (step_preserves_halting _ _ tSt'). assumption.
    intros s Rs. eapply IHT2.
    eapply T_App. apply Ht. apply R_typable_empty. apply Rs.
    apply ST_App1. apply tSt'.
    apply RRt'. apply Rs.
Case "TProd".
    split. apply (step_preserves_halting _ _ tSt'). assumption.
    destruct RRt' as [t1 [t2 [STM extra]]].
    exists t1. 
    exists t2. 
    split. 
    eapply rtc_step. apply tSt'. 
    auto. auto.
Case "TSum".
    split. apply (step_preserves_halting _ _ tSt'). assumption.
    destruct RRt' as [t1 [v_t1 STM]].
    exists t1; split; auto.
    destruct STM as [H | H]; eauto; [left | right]; inversion H; split; try assumption.
    SCase "tinl".
      inversion H0. subst. 
      SSCase "0-step".
         apply rtc_step with (tinl T2 t1); assumption.
      SSCase "> 0-step". 
         assert (t ~~>* t') as tMSt'.
            subst. apply rtc_step with t'. assumption. apply rtc_refl.
         apply rtc_Trans with t'; assumption; assumption.
    SCase "tinr".
      inversion H0. subst.
      SSCase "0-step".
         apply rtc_step with (tinr T1 t1); assumption.
      SSCase "> 0-step".
           assert (t ~~>* t') as tMSt'.
            subst. apply rtc_step with t'. assumption. apply rtc_refl.
            apply rtc_Trans with t'; assumption; assumption.
Case "TNil".
    split. apply (step_preserves_halting _ _ tSt'). assumption. 
    apply RRt'.
Case "TMonadic".
    fold Rsto. fold Rsto in RRt'. split. apply (step_preserves_halting _ _ tSt'). assumption. 
    destruct m; intros. 
    SCase "reactive".
      destruct (RRt' Sto H) as [t'0 [Sto' ?]].
      assert ((t,Sto) ~~~>* (t',Sto)) by (eapply STM_LC in tSt'; eauto). 
      exists t'0; exists Sto'. 
      destruct H0 as [? [? ?]]; auto.
    SCase "nonreactive".
      intros. 
      destruct (RRt' Sto H) as [t'' [Sto0 [? [? [? ?]]]]].
      assert ((t,Sto) ~~~>* (t',Sto)) by (eapply STM_LC in tSt'; eauto).
      exists t''; exists Sto0. 
             repeat (split; eauto).
Qed.

Lemma multistep_preserves_R' : forall T t t', 
      \empty |-- t \in T -> 
             (t ~~>* t') -> 
                R T t' -> 
                R T t.
Proof. 
  intros T t t' HT STM. 
  induction STM; intros. 
    assumption.
    eapply step_preserves_R'.  assumption. apply H. apply IHSTM. 
    eapply preservation;  eauto. auto. 
Qed.

(* ###################################################################### *)
(** *** Definition pertaining to multi-substitutions, enviroments, type assignments,
        extension of a context thereby, closed enviroments, etc. Again, fairly
        standard machinery borrowed from SF. *)

Definition env := list (id * tm). 

Fixpoint closed_env (env:env) {struct env} :=
         match env with
         | nil         => True
         | (x,t)::env' => closed t /\ closed_env env'
         end.

Fixpoint msubst (ss:env) (t:tm) {struct ss} : tm :=
         match ss with
         | nil => t
         | ((x,s)::ss') => msubst ss' ([x:=s]t)
         end.

Definition tass := list (id * Ty).

Fixpoint mextend (Gamma : context) (xts : tass) :=
         match xts with
         | nil => Gamma
         | ((x,v)::xts') => extend (mextend Gamma xts') x v
         end.

Fixpoint lookup {X:Set} (k : id) (l : list (id * X)) {struct l} : option X :=
         match l with
         | nil => None
         | (j,x) :: l' => if eq_id_dec j k then Some x else lookup k l'
         end.

Fixpoint drop {X:Set} (n:id) (nxs:list (id * X)) {struct nxs} : list (id * X) :=
         match nxs with
         | nil => nil
         | ((n',x)::nxs') => if eq_id_dec n' n then drop n nxs' else (n',x)::(drop n nxs')
         end.

Inductive instantiation :  tass -> env -> Prop :=
  | V_nil  : instantiation nil nil 
  | V_cons : forall x T v c e, 
                       value v -> 
                         R T v -> 
             instantiation c e -> 
             instantiation ((x,T)::c) ((x,v)::e).

(* ###################################################################### *)
(** *** Lemmas concerning various properties of the functions defined above. *)

Lemma mextend_lookup : forall (c:tass) (x:id), 
      lookup x c = (mextend empty c) x. 
Proof.
  induction c; intros. 
    auto.
    destruct a. unfold lookup, mextend, extend. destruct (eq_id_dec i x); auto.  
Qed.

Lemma mextend_drop : forall (c:tass) Gamma x x', 
      mextend Gamma (drop x c) x' = if eq_id_dec x x' then Gamma x' else mextend Gamma c x'.
Proof. 
  induction c; intros. 
      destruct (eq_id_dec x x'); auto. 
      destruct a. simpl.
      destruct (eq_id_dec i x).
         subst. rewrite IHc. 
            destruct (eq_id_dec x x').  auto. unfold extend. rewrite neq_id; auto. 
         simpl. unfold extend.  destruct (eq_id_dec i x').
            subst.
               destruct (eq_id_dec x x'). 
                  subst. exfalso. auto. 
                  auto. 
           auto. 
Qed.

(* ###################################################################### *)
(** *** Properties of Instantiations *)

Lemma instantiation_domains_match: forall {c} {e}, 
      instantiation c e -> 
      forall {x} {T}, lookup x c = Some T -> exists t, lookup x e = Some t.
Proof.
  intros c e V. induction V; intros x0 T0 C.
    inversion C. 
    simpl in *.
    destruct (eq_id_dec x x0); eauto. 
Qed.

Lemma instantiation_env_closed : forall c e,  
      instantiation c e -> closed_env e.
Proof.
  intros c e V; induction V; intros. 
    econstructor. 
    unfold closed_env. fold closed_env. 
    split.  eapply typable_empty_closed. eapply R_typable_empty. eauto.   
        auto.
Qed.

Lemma instantiation_R : forall c e, 
      instantiation c e -> 
                        forall x t T, 
      lookup x c = Some T -> 
      lookup x e = Some t -> 
                  R T t.
Proof.
  intros c e V. induction V; intros x' t' T' G E. 
    inversion E. 
    unfold lookup in *.  destruct (eq_id_dec x x'). 
      inversion G; inversion E; subst.  auto.
      eauto. 
Qed.

Lemma instantiation_drop : forall c env, 
  instantiation c env -> forall x, instantiation (drop x c) (drop x env).
Proof.
  intros c e V. induction V.  
    intros.  simpl.  constructor.  
    intros. unfold drop. destruct (eq_id_dec x x0); auto. constructor; eauto. 
Qed.

Lemma mextend_empty_lookup : forall c x, (mextend empty c) x = lookup x c.
Proof. 
  intros; rewrite mextend_lookup; reflexivity. 
Qed.

Lemma msubst_closed: forall t, 
              closed t -> 
                    forall ss, 
                      msubst ss t = t.
Proof. 
  induction ss as [|x]; intros; [reflexivity|destruct x; simpl].
    rewrite subst_closed; assumption.
Qed.

Lemma msubst_preserves_typing : forall c e,
     instantiation c e -> 
     forall Gamma t S, (mextend Gamma c) |-- t \in S -> 
     Gamma |-- (msubst e t) \in S.
Proof.
    induction 1; intros.  
    simpl in H. simpl. auto. 
    simpl in H2.  simpl. 
    apply IHinstantiation.
    eapply subst_preserves_typing; eauto. 
    apply R_typable_empty; auto.
Qed.

Lemma subst_msubst: forall env x v t, 
            closed v -> 
      closed_env env ->
      msubst env ([x:=v]t) = [x:=v](msubst (drop x env) t).
Proof. 
  induction env0; intros. 
    auto. 
    destruct a. simpl. 
    inversion H0. fold closed_env in H2.
    destruct (eq_id_dec i x).
      subst. rewrite duplicate_subst; auto. 
      simpl. rewrite swap_subst; eauto. 
Qed.

Lemma msubst_var:  forall ss x, closed_env ss -> 
      msubst ss (tvar x) = 
      match lookup x ss with
      | Some t => t
      | None => tvar x
      end.
Proof.
  induction ss as [|x]; intros; [reflexivity|destruct x; simpl].
      destruct (eq_id_dec i x0).
      apply msubst_closed. inversion H; auto. 
      apply IHss. inversion H; auto. 
Qed.


Lemma msubst_abs: forall ss x T t,  
      msubst ss (tabs x T t) = tabs x T (msubst (drop x ss) t).
Proof.
  induction ss; intros.
    reflexivity. 
    destruct a. 
      simpl. destruct (eq_id_dec i x); simpl; auto.
Qed.

Lemma msubst_app : forall ss t1 t2, 
      msubst ss (tapp t1 t2) = tapp (msubst ss t1) (msubst ss t2).
Proof.
 induction ss; intros.
   reflexivity.
   destruct a.  
    simpl. rewrite <- IHss. auto. 
Qed.

Lemma msubst_pair : forall ss t1 t2,
      msubst ss (tpair t1 t2) = tpair (msubst ss t1) (msubst ss t2).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_proj : forall ss t1 t2,
      msubst ss (tproj t1 t2) = tproj (msubst ss t1) (msubst ss t2).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_tinl : forall ss T t,
      msubst ss (tinl T t) = tinl T (msubst ss t).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_tinr : forall ss T t,
      msubst ss (tinr T t) = tinr T (msubst ss t).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_tcase : forall ss t1 t2 t3, 
      msubst ss (tcase t1 t2 t3) = tcase (msubst ss t1) (msubst ss t2) (msubst ss t3).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_return : forall ss M t,
      msubst ss (treturn M t) = treturn M (msubst ss t).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_bind : forall ss T t1 t2,
      msubst ss (tbind T t1 t2) = tbind T (msubst ss t1) (msubst ss t2).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_lift : forall ss M t,
      msubst ss (tlift M t) = tlift M (msubst ss t).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_elevate : forall ss M t,
      msubst ss (televate M t) = televate M (msubst ss t).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_tunit : forall ss,
      msubst ss tunit = tunit.
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite IHss. auto.
Qed.

Lemma msubst_get : forall ss M,
      msubst ss (tget M) = tget M.
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite IHss. auto.
Qed.

Lemma msubst_tput : forall ss t M,
      msubst ss (tput M t) = tput M (msubst ss t).
Proof.
  induction ss; intros.
    reflexivity.
    destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_trunst : forall ss t1 t2,
                        msubst ss (trunst t1 t2) = trunst (msubst ss t1) (msubst ss t2).
Proof.
  induction ss; intros.
  simpl. reflexivity.
      destruct a.
      simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_trunid : forall ss t,
                        msubst ss (trunid t) = trunid (msubst ss t).
Proof.
  induction ss; intros.
    simpl. reflexivity.
        destruct a.
        simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_tpause : forall ss t M T,
                         msubst ss (tpause M T t) = tpause M T (msubst ss t).
Proof.
  induction ss; intros.
     reflexivity.
     destruct a.
     simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_tunfold : forall ss t1 t2 TA TB M,
                         msubst ss (tunfold M TA TB t1 t2) = tunfold M TA TB (msubst ss t1) (msubst ss t2).
Proof.
  induction ss as [|x]; intros.
     reflexivity.
     destruct x.
     simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_trunre : forall ss t T,
                        msubst ss (trunre T t) = trunre T (msubst ss t).
Proof.
  induction ss; intros.
  reflexivity.
  destruct a.
  simpl. rewrite <- IHss. auto.
Qed.

Lemma msubst_rewrite : forall ss t,
      msubst ss t = (match t with
                     | (tabs x T t')           => tabs x T (msubst (drop x ss) t')
                     | (tapp t1 t2)            => tapp (msubst ss t1) (msubst ss t2)
                     | (tpair t1 t2)           => tpair (msubst ss t1) (msubst ss t2)
                     | (tproj t1 t2)           => tproj (msubst ss t1) (msubst ss t2)
                     | (tinl T t')             => tinl T (msubst ss t')
                     | (tinr T t')             => tinr T (msubst ss t')
                     | (tcase t1 t2 t3)        => tcase (msubst ss t1) (msubst ss t2) (msubst ss t3)
                     | (treturn M t')          => treturn M (msubst ss t')
                     | (tbind T t1 t2)         => tbind T (msubst ss t1) (msubst ss t2)
                     | (tlift M t')            => tlift M (msubst ss t')
                     | (televate M t')         => televate M (msubst ss t')
                     |  tunit                  => tunit
                     | (tget M)                => tget M
                     | (tput M t')             => tput M (msubst ss t')
                     | (trunst t1 t2)          => trunst (msubst ss t1) (msubst ss t2)
                     | (trunid t')             => trunid (msubst ss t')
                     | (tpause M T t')         => tpause M T (msubst ss t')
                     | (tunfold M TA TB t1 t2) => tunfold M TA TB (msubst ss t1) (msubst ss t2)
                     | (trunre T t')           => trunre T (msubst ss t')
                     | _                       => msubst ss t
                     end).
Proof. 
 induction ss as [|x]; 
 [induction t;auto|induction t;destruct x;simpl;auto]; rewrite IHss; auto.
 destruct (eq_id_dec i0 i); simpl; auto.
Qed.

(** Specialzed tactic for working with msubst in our proof of the substitution lemma: *) 

Ltac rewrite_msubst := 
     match goal with 
     | [ |- appcontext[msubst ?ss ?t] ] => 
       match t with
       | (tvar ?x)  => (match goal with 
                       | [H1: (?Gamma ?x) = (Some ?T), 
                          H2: (forall x0, (?Gamma x0) = (lookup x0 ?c)), 
                          H3: (instantiation ?c ?env) |- _ ] => 
                              rewrite H2 in H1; 
                              let TEMP := fresh in
                                  destruct (instantiation_domains_match H3 H1) as [? TEMP];
                                  pose proof (instantiation_R c env H3 _ _ T H1 TEMP);
                                  pose proof (instantiation_env_closed _ _ H3);
                                  rewrite msubst_var; [rewrite TEMP|];auto
                       end)
       | _          => rewrite msubst_rewrite; auto
       end
     end.



(* ###################################################################### *)
(** *** The following lemmas arise out of an unavoidable use of the COFIX tactic, the use of 
        which is (unfortunately) required in order to prove the tbind and tunfold cases of the
        substitution lemma. *)

Lemma step_mo_still_values_star : forall t t' Sto Sto',
                               store_all_values Sto ->
                               (t,Sto) ~~~>* (t',Sto') ->
                               store_all_values Sto'.
Proof.
  intros ? ? ? ? Hav Hstep.
  remember (t,Sto) as co.
  remember (t',Sto') as co'.
  generalize dependent t Sto t' Sto'.
  induction Hstep; intros; subst.
    inv Heqco'. 
    destruct y as (u,uSto). 
    pose proof (step_mo_still_values _ _ _ _ Hav H). 
    apply (IHHstep _ _ H0 eq_refl _ _ eq_refl).
Qed. 

Lemma tbind_along : forall t Sto SM TI TO T1 T2 t1 t2 Sto' Sto'',
                      \empty |-- t \in TMonadic (MReactT TI TO SM) T2 ->
                      store_all_values Sto ->
                      store_matches_mo Sto (MReactT TI TO SM) ->
                      (t,Sto) ~~~>* (tbind T2 t1 t2,Sto') ->
                      \empty |-- t1 \in TMonadic (MReactT TI TO SM) T1 ->
                      store_all_values Sto' ->
                      store_matches_mo Sto' (MReactT TI TO SM) ->
                      \empty |-- t2 \in TArrow T1 (TMonadic (MReactT TI TO SM) T2) ->
                      value t1 ->
                      value t2 ->
                      Rsto SM Sto'' ->
                      along_react (Rsto SM) (R TI) (R TO) (R T1) (t1,Sto') ->
                      (forall t' Sto'',
                         R T1 t' ->
                         Rsto SM Sto'' ->
                         along_react (Rsto SM) (R TI) (R TO) (R T2) ((tapp t2 t'), Sto'')) ->
                      along_react (Rsto SM) (R TI) (R TO) (R T2) (t,Sto).
Proof. cofix; intros.
  inversion H10; subst.
  Case "treturn".
    pose proof (H11 _ _ H20 H21).
    eapply step_preserves_along_star.
    eapply H2.
    eapply step_preserves_along_star.
    eapply Congruence_Mo_Bind1; eauto.
    eapply step_preserves_along.
    eapply STM_BindRet; eauto.    
    assumption.
  Case "tpause".
    assert (TI = TI0 /\ TO = TO0 /\ SM = SM0 /\ T = T1).
           assert ((t1, Sto') |> TMonadic (MReactT TI TO SM) T1). 
                  constructor. auto. auto. auto.
           assert ((tpause (MReactT TI0 TO0 SM0) T t', Sto'0) |> TMonadic (MReactT TI TO SM) T1).
                  eapply preservation_mo_star;eauto.       
           inversion H13.
           inversion H28.
           auto.
    destruct H12 as [HRe1 [HRe2 [HRe3 HRe4]]].
             rewrite <- HRe1 in *; 
             rewrite <- HRe2 in *; 
             rewrite <- HRe3 in *; 
             rewrite <- HRe4 in *; clear HRe1 HRe2 HRe3 HRe4.
    assert (\empty |-- t' \in TMonadic (MNonReact SM) (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
    assert ((tpause (MReactT TI TO SM) T t', Sto'0) |> TMonadic (MReactT TI TO SM) T).      
           eapply preservation_mo_star;eauto.
           inversion H12.
           inversion H25.
           auto.
    assert (HTvlr : \empty |-- vl \in TO /\
            \empty |-- vr \in TArrow TI (TMonadic (MReactT TI TO SM) T)).
    assert ((treturn (MNonReact SM) (tpair vl vr), Sto''0) |> TMonadic (MNonReact SM)
             (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
    assert ((t',Sto'0) |> TMonadic (MNonReact SM)
                        (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
           constructor; auto. 
           apply (step_mo_still_values_star _ _ _ _ H4 H14).
           eapply Rsto_matches;eauto.
           eapply preservation_mo_star;eauto.
     inversion H13.
     inversion H28.
     inversion H33.
     auto.
     destruct HTvlr as [HTvl HTvr].
     assert ((t1 >>=[T2] t2, Sto') ~~~>* (tpause (MReactT TI TO SM) T t' >>=[T2] t2, Sto'0)) by (eapply Congruence_Mo_Bind1; eauto).
     pose proof (rtc_Trans_mo _ _ _ H2
                (rtc_Trans_mo _ _ _ H13
                              (Congruence_STM_PauseBind t' t2 Sto'0 TI TO SM T T2 H15 H8))).
     assert (HGoal: (t' >>=[TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T2))] 
          (\(0:TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))) 
            tproj (tvar 0)
              (\(1:TO) (\(2:TArrow TI (TMonadic (MReactT TI TO SM) T)) 
                         treturn (MNonReact SM)
                           (tpair (tvar 1)
                              (\(3:TI) (tapp (tvar 2) (tvar 3) >>=[T2] t2)))))),
        Sto'0) ~~~>*
        (treturn (MNonReact SM)
                 (tpair vl (\(3:TI) (tapp vr (tvar 3) >>=[T2] t2))), Sto''0)).
            rtc_Trans.
               eapply Congruence_Mo_Bind1.
                      constructor.
                      eapply H20.
            rtc_Trans.
               eapply Congruence_BindRet.
                      constructor;auto.
                      constructor.
            rtc_Trans. 
               eapply Congruence_AppAbs_Mo.
                      constructor;auto.
                      simpl.
                      rewrite_subst.
            rtc_Trans.          
               eapply Congruence_LC.
                      eapply rtc_R.
                             eapply ST_Proj; eauto.
            rtc_Trans.
               eapply Congruence_LC. 
               eapply Congruence_App1.
                      eapply rtc_R.
                             eapply ST_AppAbs.
                             auto.
                             simpl.
                             rewrite_subst.
            rtc_Trans.
               eapply Congruence_AppAbs_Mo.
                      auto.
                      simpl.
                      rewrite_subst.
            eauto.
     assert (value (\(3:TI) (tapp vr (tvar 3) >>=[T2] t2))) by constructor.                 
     eapply along_pause.     
            eapply H22.
            constructor.
            auto.
            constructor.
            eapply H16.
            eapply H23.
            auto.
            eapply H19.
            auto.
            auto.
    SCase "halts".
       intros.
       destruct (R_halts H24) as [v'' [? ?]].
       pose proof (multistep_preserves_R _ _ _ H25 H24).
       pose proof (Congruence_App2 _ _ _ H23 H25).
       pose proof (rtc_R _ _ (ST_AppAbs 3 TI (tapp vr (tvar 3) >>=[T2] t2) v'' H28)).
       simpl in H31. rewrite_subst.
       destruct (H26 _ H29) as [vapp []].
       assert ((tapp vr v'' >>=[T2] t2) ~~>* (vapp >>=[T2] t2)) by (eapply Congruence_Tbind1; eauto).
       pose proof (rtc_Trans _ _ _ H30
                  (rtc_Trans _ _ _ H31 H34)).
       eapply multistep_preserves_halting.
       eapply H35.
       apply values_halt.
             constructor;auto.
     SCase "along".       
       intros.
       destruct (R_halts H24) as [v'' [? ?]].
       pose proof (multistep_preserves_R _ _ _ H28 H24).
       pose proof (Congruence_App2 _ _ _ H23 H28).
       pose proof (rtc_R _ _ (ST_AppAbs 3 TI (tapp vr (tvar 3) >>=[T2] t2) v'' H29)).
       simpl in H32. rewrite_subst.
       destruct (H26 _ H30) as [vapp []].
       assert ((tapp vr v'' >>=[T2] t2) ~~>* (vapp >>=[T2] t2)) by (eapply Congruence_Tbind1; eauto).
       pose proof (Congruence_LC _ _ Sto''1
                  (rtc_Trans _ _ _ H31
                  (rtc_Trans _ _ _ H32 H35))).
       pose proof (H27 _ _ H30 H25).
       pose proof (Congruence_LC _ _ Sto''1 H33).
       eapply tbind_along; eauto.
              repeat econstructor;eauto.
              eapply typable_empty_everywhere.
              eauto.
              weakening.
              apply (R_typable_empty H24).
              eapply Rsto_all_values; eauto.
              constructor.
              eapply Rsto_matches;auto.
              eapply preservation_star;eauto.
              econstructor.
              eauto.
              apply (R_typable_empty H30).
              eapply Rsto_all_values; eauto.
              constructor.
              eapply Rsto_matches;auto.
              eapply step_preserves_along_star'.
              eapply H38.
              auto.
Qed.       

Lemma tunfold_along : forall t Sto TA TB TI TO SM t1 t2 Sto',
                        ((t,Sto) ~~~>* (tunfold (MReactT TI TO SM) TA TB t1 t2,Sto')) ->
                        R TB t1 ->
                        R (TArrow TB (TMonadic (MNonReact SM) (TSum TA (TProd TO (TArrow TI TB))))) t2 ->
                        Rsto SM Sto' ->
                        along_react (Rsto SM) (R TI) (R TO) (R TA) (t,Sto).
Proof.
  cofix; intros ? ? ? ? ? ? ? ? ? ? Hstep HRt1 HRt2 HRSto.
  destruct (R_halts HRt1) as [vt1 [? ?]].
  destruct (R_halts HRt2) as [vt2 [? ?]].
  pose proof (multistep_preserves_R _ _ _ H HRt1).
  pose proof (multistep_preserves_R _ _ _ H1 HRt2).
  destruct H4 as [? [? ?]].
  pose proof (H6 _ H3).
  destruct H7 as [? [? ?]].
  pose proof (H9 _ HRSto).
  destruct H10 as [vret [Sto'' [? [? [? ?]]]]].
  destruct H12 as [? [? ?]].
  destruct H15 as [vinner [? ?]].
  destruct H16.
  SCase "vret goes left".
    destruct H8 as [vapp [? ?]].
    destruct H16.
    eapply step_preserves_along_star.
    eapply Hstep.
    eapply step_preserves_along_star.
    eapply Congruence_LC.
    eapply Congruence_Unfold1; eauto.
    eapply step_preserves_along_star.
    eapply Congruence_LC.
    eapply Congruence_Unfold2; eauto.
    eapply step_preserves_along.
    eapply STM_Unfold; eauto.
    eapply step_preserves_along_star.
    eapply Congruence_Mo_Bind1.
      constructor.
    eapply Congruence_LC.
    eapply Congruence_Lift.
    eapply H8.
    assert (done_mo (treturn (MNonReact SM) vret,Sto'')) by (constructor; auto).
    pose proof (step_val_done _ _ _ _ _ H8 H17 H10 H19).
    eapply step_preserves_along_star.
    eapply Congruence_Mo_Bind1.
      constructor.
      eapply Congruence_LiftRe; eauto.
    eapply step_preserves_along_star.
      eapply Congruence_Mo_Bind1.
      constructor.
      eapply rtc_R_mo.
      eapply STM_LiftRetRe; eauto.
    eapply step_preserves_along.
      eapply STM_BindRet; eauto.
    eapply step_preserves_along.
      eapply STM_LC.
      eapply ST_AppAbs; eauto.
      simpl.
      rewrite_subst.
    eapply step_preserves_along_star.
      eapply Congruence_LC.
      eapply Congruence_Tcase; eauto.
    eapply step_preserves_along.
      eapply STM_LC.
        eapply ST_CaseL; eauto.
    eapply step_preserves_along.
      eapply STM_LC.
      eapply ST_AppAbs; eauto.
      simpl.
    eapply along_return.
      eapply rtc_refl.
      auto.
      auto.
      auto.
  SCase "vret goes right".
    destruct H8 as [vapp [? ?]].
    destruct H16.
    destruct H18 as [? [? ?]].
    destruct H20 as [vL [vR [? [? [? [? ?]]]]]].
    assert (done_mo (treturn (MNonReact SM) vret,Sto'')) by (constructor; auto).
    pose proof (step_val_done _ _ _ _ _ H8 H17 H10 H25). 
    eapply along_pause. 
    eapply rtc_Trans'.
    eapply Hstep.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply Congruence_Unfold1; eauto.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply Congruence_Unfold2; eauto.
    eapply rtc_Trans'.
    eapply rtc_R'.
    eapply STM_Unfold; eauto.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply Congruence_Tbind1.
    eapply Congruence_Lift; eauto.
    eapply rtc_Trans'.
    eapply Congruence_Mo_Bind1.
      constructor.
    eapply Congruence_LiftRe; eauto.
    eapply rtc_Trans'.
    eapply Congruence_Mo_Bind1.
      constructor.
    eapply rtc_R'.
    eapply STM_LiftRetRe; eauto.
    eapply rtc_Trans'.
    eapply rtc_R'.
    eapply STM_BindRet; eauto.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply rtc_R.
    eapply ST_AppAbs; eauto.
    simpl.
      rewrite_subst.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply Congruence_Tcase; eauto.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply rtc_R'.
    eapply ST_CaseR; eauto.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply rtc_R.
    eapply ST_AppAbs; eauto.
    simpl.
      rewrite_subst.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply Congruence_Proj1; eauto.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply rtc_R.
    eapply ST_Proj; eauto.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply Congruence_App1.
    eapply rtc_R.
    eapply ST_AppAbs; eauto.
    simpl.
      rewrite_subst.
    eapply rtc_Trans'.
    eapply Congruence_LC.
    eapply rtc_R.
    eapply ST_AppAbs; eauto.
    pose proof (R_typable_empty H23).
    simpl.
      rewrite_subst.
    apply rtc_refl.
    constructor.
    constructor.
    auto.
    constructor. eapply H21.
    instantiate (1:= (\(4:TI) tunfold (MReactT TI TO SM) TA TB (tapp vR (tvar 4)) vt2)).
    constructor.
    auto.
    eapply H13.
    eauto.
    auto.
    intros.
    destruct (R_halts H27) as [vt''' [? ?]].
    pose proof (preservation_star TI _ _ H28 (R_typable_empty H27)).
    pose proof (preservation_star (TProd TO (TArrow TI TB)) _ _ H20 H18).
    inv H31.
    eapply multistep_preserves_halting.
    eapply Congruence_App2.
      constructor.
      eapply H28.
    eapply step_preserves_halting.
    eapply ST_AppAbs; eauto.    
    simpl.
      rewrite_subst.
    destruct H24 as [? [? ?]].
    pose proof (multistep_preserves_R _ _ _ H28 H27).
    pose proof (H33 _ H34).
    destruct (R_halts H35) as [vappR [? ?]].
    eapply multistep_preserves_halting.
    eapply Congruence_Unfold1; eauto.
    apply values_halt; repeat constructor; auto.
    intros.
    destruct (R_halts H27) as [vt''' [? ?]].
    pose proof (preservation_star TI _ _ H29 (R_typable_empty H27)).
    pose proof (preservation_star (TProd TO (TArrow TI TB)) _ _ H20 H18).
    inv H32.
    eapply tunfold_along.
    eapply rtc_Trans'.
      eapply Congruence_LC.
      eapply Congruence_App2.
        constructor.
        eapply H29.
    eapply rtc_Trans'.
      eapply rtc_R'.
      eapply STM_LC.
      eapply ST_AppAbs.
    assumption.      
    simpl.
      rewrite_subst.
    eapply rtc_refl.
    destruct H24 as [? [? ?]].
    apply H34.
    eapply multistep_preserves_R; eauto.
    eapply multistep_preserves_R;eauto.
    assumption.
Qed.    

Lemma tbind_R : forall M T1 T2 t t',
  R (TMonadic M T1) t ->
  R (TArrow T1 (TMonadic M T2)) t' ->
  R (TMonadic M T2) (tbind T2 t t').
Proof.
  intros.
  pose proof (R_typable_empty H) as HHt;pose proof (R_typable_empty H0) as HHt';
  destruct (R_halts H) as [vt [HStv Hvalt]]; destruct (R_halts H0) as [vt' [HSt'v Hvalt']];
  assert (HStepttv: tbind T2 t t' ~~>* tbind T2 vt vt') by (pose proof (Congruence_Tbind1 T2 _ _ t' HStv);
                                                      pose proof (Congruence_Tbind2 T2 _ _ _ Hvalt HSt'v); eauto).
  assert (halts (tbind T2 vt vt')) by (apply values_halt; constructor; auto).
    unfold R; fold Rsto; fold R.
  assert (WT: \empty |-- tbind T2 t t' \in TMonadic M T2) by (econstructor;eauto).  
  split; [ assumption | split; [eapply multistep_preserves_halting;eauto| ]].
  clear H1.
    pose proof (multistep_preserves_R _ _ _ HStv H).
    pose proof (multistep_preserves_R _ _ _ HSt'v H0).
  Case "M".
  destruct M; intros.
  pose proof (Rsto_matches _ _ H3).
  assert (HStoM: store_matches_mo Sto (MReactT t0 t1 s)) by (constructor; auto).
  assert (WTco : (tbind T2 t t',Sto) |> (TMonadic (MReactT t0 t1 s) T2)). 
         constructor. auto. 
                      eapply Rsto_all_values;eauto. auto. 
  Focus 2.
  SCase "nonreactive".
        destruct H as [? [? ?]].
        destruct H0 as [? [? ?]].
        pose proof (Congruence_LC _ _ Sto HStepttv).
        destruct H1 as [? [? ?]].
        destruct H2 as [? [? ?]].
        pose proof (H10 _ H3).
        destruct H13 as [t'0 [Sto'0 [? [? [? ?]]]]].
        pose proof (H10 _ H16).
        destruct H17 as [? [? ?]].
        pose proof (H12 _ H15). 
        unfold R in H18; fold Rsto in H18; fold R in H18.
        destruct H18 as [? [? ?]].
        pose proof (H20 _ H16).
        pose proof (Congruence_Mo_Bind1 T2 vt Sto _ Sto'0 vt' Hvalt' H13).
        pose proof (STM_BindRet _ _ T2 (MNonReact s) Sto'0 H14 Hvalt').
        destruct H21 as [t_m [Sto_m [? [? [? ?]]]]].
        assert ((tbind T2 t t', Sto) ~~~>* (treturn (MNonReact s) t_m, Sto_m)).
               eapply rtc_Trans_mo.
               eapply H8.
               eapply rtc_Trans_mo.
               eapply H22.
               apply rtc_R_mo in H23.
               eapply rtc_Trans_mo.
               eapply H23.
               auto.
         exists t_m; exists Sto_m.
         repeat (split; auto).
   SCase "Reactive".  
     destruct H1 as [? [? ?]].
     destruct H5 as [vvt [? ?]].
     pose proof (H6 _ H3).
     destruct H2 as [? [? ?]].
     destruct H8 as [t'0 [Sto' [? [? [? ?]]]]].
     inversion H13; subst.
     SSCase "going to return".
       pose proof (H10 _ H22).
       destruct H14 as [? [? ?]]. 
       assert ((vt, Sto) ~~~>* (treturn (MReactT TI TO SM) t'1, Sto'0)) by (eauto).
       pose proof (Congruence_Mo_Bind1 T2 _ Sto _ Sto'0 vt' Hvalt' H19).
       pose proof (Congruence_BindRet T2 t'1 vt' (MReactT TI TO SM) Sto'0 H17 Hvalt').
       assert ((tbind T2 t t', Sto) ~~~>* (tapp vt' t'1, Sto'0)).
              eapply rtc_Trans_mo.
              instantiate (1:= (tbind T2 vt vt',Sto)).
              eapply Congruence_LC; eauto.
              eapply rtc_Trans_mo.
              eapply H20.
              auto.
        pose proof (H18 _ H23).
        destruct H25 as [t'2 [Sto'1 [? [? [? ?]]]]].
        exists t'2, Sto'1.
        repeat (split;eauto).
    SSCase "going to pause".
     assert (HRewrite: TI = t0 /\ TO = t1 /\ SM = s /\ T = T1 /\  \empty |-- t'1 \in
        TMonadic (MNonReact s)
          (TProd t1 (TArrow t0 (TMonadic (MReactT t0 t1 s) T1)))).    
     assert ((t'0, Sto') |> TMonadic (MReactT t0 t1 s) T1).
            eapply preservation_mo_star.
            eapply H8.
            constructor.
            eapply preservation_star;eauto. 
            eapply Rsto_all_values;eauto.
            constructor.
            eapply Rsto_matches;eauto.
     assert ((tpause (MReactT TI TO SM) T t'1, Sto'0) |> TMonadic (MReactT t0 t1 s) T1).
                  eapply preservation_mo_star;eauto.       
     inversion H15.
     inversion H30.
     auto.
     destruct HRewrite as [HRe1 [HRe2 [HRe3 [HRe4 HTt'1]]]].
       rewrite <- HRe1 in *; 
       rewrite <- HRe2 in *;
       rewrite <- HRe3 in *; 
       rewrite <- HRe4 in *; clear HRe1 HRe2 HRe3 HRe4.
     assert (HTvlr : \empty |-- vl \in TO /\ \empty |-- vr \in TArrow TI (TMonadic (MReactT TI TO SM) T)).
            assert ((treturn (MNonReact SM) (tpair vl vr), Sto'') |> TMonadic (MNonReact SM)
                                                            (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
                   eapply preservation_mo_star;eauto.     
                   constructor;auto. 
                   eapply Rsto_all_values;eauto.
                   eapply Rsto_matches;auto.
            inversion H14.
            inversion H27.
            inversion H34.
            auto.
     destruct HTvlr as [HTvl HTvr].
     assert (H8': (vt,Sto) ~~~>* (tpause (MReactT TI TO SM) T t'1,Sto'0)) by eauto.       
     assert ((vt >>=[T2] vt', Sto) ~~~>* (tpause (MReactT TI TO SM) T t'1 >>=[T2] vt', Sto'0)) by (eapply Congruence_Mo_Bind1; eauto).
     exists (vt >>=[T2] vt'), Sto.
     split;[eapply Congruence_LC; eauto|
             split;[constructor;  auto |
                     split;[      auto | ]]].
           eapply tbind_along.
      econstructor;eauto.
      eapply Rsto_all_values;eauto.
      constructor.
      eapply Rsto_matches;eauto.
      eapply H14.
      econstructor;eauto.
      eapply Rsto_all_values;eauto.
      constructor.
      eapply Rsto_matches;eauto.
      auto.
      constructor;auto.
      auto.
      eassumption.
      eapply step_preserves_along_star'.
      eapply H16.
      auto.
      intros.
         pose proof (H10 _ H15).
         destruct H25 as [? [? ?]].
         destruct (H27 _ H24) as [t'3 [Sto'1 [? [? [? ?]]]]]. 
         eapply step_preserves_along_star;eauto.
Qed.

(* ###################################################################### *)
(** *** From msubst_R it follows that the ReWire calculus is strongly normalizing, 
        for both lambda and monadic expressions. This is because in order to be 
        a member of their respective [R T]'s, a term must be well-typed, and either 
        have or possess the property of halting. This lemma is more commonly
        referred to as the 'Substitution Lemma' in textbooks and whatnot. *)

Lemma msubst_R : forall c env t T, 
     (mextend empty c) |-- t \in T -> 
     instantiation c env -> 
     R T (msubst env t).
Proof.
  intros c env0 t T HT V; 
         generalize dependent env0;
         remember (mextend empty c) as Gamma;
         generalize (mextend_empty_lookup c) as H; intros H; rewrite <- HeqGamma in H;
         clear HeqGamma;generalize dependent c;
  induction HT; intros; rewrite_msubst. 
  Case "T_Abs". 
     assert (WT:  empty |-- (tabs x T (msubst (drop x env0) t)) \in (TArrow T T')).
     eapply T_Abs. eapply msubst_preserves_typing.  eapply instantiation_drop; eauto.  
      eapply context_invariance.  apply HT.  
      intros. 
      unfold extend. rewrite mextend_drop. destruct (eq_id_dec x x0). auto.
        rewrite H.  
          clear - c n. induction c. 
              simpl. rewrite neq_id; auto. 
              simpl. destruct a.  unfold extend. destruct (eq_id_dec i x0); auto. 
    unfold R. fold R. split. 
       auto. 
     split. apply values_halt. apply v_abs.
     intros.
     destruct (R_halts H0) as [v [P Q]].
     pose proof (multistep_preserves_R _ _ _ P H0).
     apply multistep_preserves_R' with (msubst ((x,v)::env0) t).
       eapply T_App. eauto. 
       apply R_typable_empty; auto. 
       eapply rtc_Trans.  eapply Congruence_App2; eauto.
       eapply rtc_R.
       simpl. rewrite subst_msubst at 1. eapply ST_AppAbs; eauto.
       eapply typable_empty_closed.
       apply (R_typable_empty H1). 
       eapply instantiation_env_closed; eauto. 
       eapply (IHHT ((x,T)::c)).  
          intros. unfold extend, lookup. destruct (eq_id_dec x x0); auto. 
       constructor; auto. 
  Case "T_App".
    destruct (IHHT1 c H env0 V) as [_ [_ P1]]. 
    pose proof (IHHT2 c H env0 V) as P2.  fold R in P1.  auto. 
  Case "TNil".
    unfold R;
    assert (value tunit) by (constructor; auto). 
    split; [ apply T_Unit |
             split; [apply values_halt in H0; assumption | auto]]. 
  Case "TProd".
    pose proof (IHHT1 c H env0 V); pose proof (IHHT2 c H env0 V).
    unfold R. fold R. split. apply T_Pair; apply R_typable_empty; auto.
    destruct (R_halts H0) as [v1' [Hstepv1' Hval_v1']]; destruct (R_halts H1) as [v2' [Hstepv2' Hval_v2']].
    pose proof (Congruence_Pair1 _ _ (msubst env0 t2) Hstepv1').
    pose proof (Congruence_Pair2 _ _ _ Hval_v1' Hstepv2').
    assert (tpair (msubst env0 t1) (msubst env0 t2) ~~>* tpair v1' v2') by (eauto).
    assert (halts (tpair v1' v2')) by (eapply values_halt; constructor; auto).
    split; [eapply multistep_preserves_halting; eauto |exists v1'; exists v2';
              repeat split; [ auto
                            | auto
                            | auto
                            | eapply multistep_preserves_R; eauto
                            | eapply multistep_preserves_R; eauto
                            ]].
 Case "T_Proj".
    pose proof (IHHT1 c H env0 V); pose proof (IHHT2 c H env0 V).
    destruct H0 as [? [? [? [? [? [? [? [? ?]]]]]]]].
    destruct H1 as [? [? ?]].
    pose proof (H9 _ H6).
    destruct H10 as [? [? ?]].
    pose proof (H12 _ H7).
    eapply multistep_preserves_R'.
      econstructor; eauto.
      eapply Congruence_Proj1; eauto.
    destruct H8 as [? [? ?]].
    eapply multistep_preserves_R'.
      econstructor; eauto. econstructor; eapply R_typable_empty; eauto.
      eapply Congruence_Proj2; eauto.
    pose proof (preservation_star _ _ _ H8 H1).
    pose proof (canonical_forms_fun _ _ _ H15 H14) as [x2 [u ?]]; subst.
    eapply multistep_preserves_R'.
      econstructor. eapply preservation_star. eapply H3. eassumption. assumption.
      eapply rtc_R.
      apply ST_Proj; auto.
    assert (Hstep: (tapp (tapp (msubst env0 t') x) x0) ~~>* (tapp (tapp (tabs x2 T1 u) x) x0)).
      eapply Congruence_App1 in H8.
      eapply Congruence_App1 in H8.
      eauto.
    eapply multistep_preserves_R; eauto.
 Case "T_Inl".
    pose proof (IHHT c H env0 V).
    destruct (R_halts H0) as [v1' [STM_v1' value_v1']].
    assert (R T1 v1') by (eapply multistep_preserves_R; eauto).
    eapply Congruence_Tinl in STM_v1'; eauto. instantiate (1 := T2) in STM_v1'.
    assert (value (tinl T2 v1')) by (constructor; auto).
    eapply R_typable_empty in H0.    
    repeat split; [econstructor; eauto
                  |eapply values_halt in H2; eapply multistep_preserves_halting; eauto
                  |exists v1';
                   split; [assumption | left; eauto ]].
 Case "T_Inr".
    pose proof (IHHT c H env0 V).
    destruct (R_halts H0) as [v1' [STM_v1' value_v1']].
    assert (R T2 v1') by (eapply multistep_preserves_R; eauto).
    eapply Congruence_Tinr in STM_v1'; eauto. instantiate (1 := T1) in STM_v1'.
    assert (value (tinr T1 v1')) by (constructor; auto).
    eapply R_typable_empty in H0.    
    repeat split; [econstructor; eauto
                  |eapply values_halt in H2; eapply multistep_preserves_halting; eauto
                  |exists v1';
                   split; [assumption | right; eauto ]].
 Case "T_Case".
    destruct (IHHT2 c H env0 V) as [HTl [H_lhalts Hlapp]];
         destruct H_lhalts as [lv [Hlstep Hlval]];
    destruct (IHHT3 c H env0 V) as [HTr [H_rhalts Hrapp]];
         destruct H_rhalts as [rv [Hrstep Hrval]];     
    destruct (IHHT1 c H env0 V) as [HTs [H_shalts Hsum]];
         destruct H_shalts as [sv [Hsstep Hsval]];
         destruct Hsum as [t' [Hvt' [[Hrl Hrt'] | [Hrr Hrt']]]].
    SCase "left".
      eapply multistep_preserves_R'.
        eapply T_Case; eauto.
        eapply Congruence_Tcase; eauto.
        eapply step_preserves_R'.
          eapply T_Case; eauto.
          eapply preservation_star; eauto.
          apply ST_CaseL; auto.
        apply Hlapp; auto.
    SCase "right".
      eapply multistep_preserves_R'.
        eapply T_Case; eauto.
        eapply Congruence_Tcase; eauto.
        eapply step_preserves_R'.
          eapply T_Case; eauto.
          eapply preservation_star; eauto.
          apply ST_CaseR; auto.
        apply Hrapp; auto.
  Case "T_Return".
    pose proof (IHHT c H env0 V).
    destruct (R_halts H0) as [v [HStv Hvalt]]. 
    assert (WT: \empty |-- treturn M (msubst env0 t) \in TMonadic M T) by (constructor; apply (R_typable_empty H0)).
    split; [ assumption
           | split; [ exists (treturn M v); split; 
                    [ eapply Congruence_LC_Ret in HStv; eauto 
                    | constructor; auto]
                    | destruct M; intros]].
    SCase "Reactive".
       exists (treturn (MReactT t0 t1 s) v); exists Sto.
       intros. 
          repeat split; [ apply (Congruence_Treturn _ _ (MReactT t0 t1 s) Sto HStv)
                        | constructor; auto 
                        | auto | ].
          eapply along_return; eauto.
          eapply multistep_preserves_R.
          eapply HStv.
       assumption.
     SCase "nonreactive".
       exists v; exists Sto.
       repeat split; [eapply Congruence_Treturn; eauto 
                     |assumption
                     |eapply multistep_preserves_R; eauto
                     |assumption ].
  Case "T_Bind".
    pose proof (IHHT1 c H env0 V);pose proof (IHHT2 c H env0 V).
    eapply tbind_R; eauto.
  Case "T_Lift St".
    pose proof (IHHT c H env0 V).
    destruct (R_halts H0) as [v [? ?]].
    pose proof (multistep_preserves_R _ _ _ H1 H0).
    destruct H3 as [? [? ?]].
    assert (HG1: tlift (MNonReact (MStateT T E SM)) (msubst env0 t) ~~>* tlift (MNonReact (MStateT T E SM)) v) by (eapply Congruence_Lift; eauto);
    split;[constructor;apply (R_typable_empty H0)
          |split;[eapply multistep_preserves_halting;[eapply Congruence_Lift;eauto
                                                     |eapply values_halt; constructor; auto]
          |intros]].
    destruct H6 as [? [? [t0 [Sto' [? [? ?]]]]]]; destruct (H5 _ H9) as [t' [Sto'' [? [? [? ?]]]]].
    assert (HGoal1 : (tlift (MNonReact (MStateT T E SM)) (msubst env0 t), Sto) ~~~>*
      (treturn (MNonReact (MStateT T E SM)) t', t0 :: Sto'')). 
            eapply rtc_Trans_mo.
                   eapply Congruence_LC.
                          eapply Congruence_Lift;eauto.
                          rewrite -> H10.
            eapply rtc_Trans_mo.
                   eapply Congruence_LiftSt.
                          auto. 
                          eapply H11.
            eapply rtc_Trans_mo. 
                   eapply Congruence_LiftRetSt.
                          auto.
                   eapply rtc_refl.
    assert (HGoal2: store_matches_mo (t0 :: Sto'') (MNonReact (MStateT T E SM))) by (
           assert (store_matches_mo Sto'' (MNonReact SM)) by (apply (Rsto_matches _ _ H14));
           constructor;[apply (R_typable_empty H8)| auto]).
    exists t'; exists (t0::Sto'').
    repeat split; auto. 
           eapply step_mo_still_values_star; eauto. 
           exists t0;exists Sto''; 
           repeat split; auto.
 Case "T_Lift Re".
    pose proof (IHHT c H env0 V);
    destruct (R_halts H0) as [v [? ?]];
    pose proof (multistep_preserves_R _ _ _ H1 H0);
    destruct H3 as [? [? ?]].
    assert (HG1: tlift (MReactT TI TO SM) (msubst env0 t) ~~>* tlift (MReactT TI TO SM) v) by (eapply Congruence_Lift;eauto).
    split;[constructor;apply (R_typable_empty H0)
          |split;[eapply multistep_preserves_halting;[eapply Congruence_Lift;eauto
                                                     |assert (halts (tlift (MReactT TI TO SM) v)) by (eapply values_halt;constructor; auto); auto]
          |intros]].
    destruct (H5 _ H6) as [t' [Sto'' [? [? [? ?]]]]]. 
    pose proof (Congruence_LiftRetRe t' TI TO SM Sto'' H8).
    pose proof (Congruence_LiftRe _ _ Sto Sto'' TI TO SM H2 H7).
    pose proof (Congruence_LC _ _ Sto HG1).
    assert (HG2 : (tlift (MReactT TI TO SM) (msubst env0 t), Sto) ~~~>* 
                  (treturn (MReactT TI TO SM) t', Sto'')) by (
                            eapply rtc_Trans_mo; eauto;
                            eapply rtc_Trans_mo; eauto).
     exists (treturn (MReactT TI TO SM) t'); exists Sto''.
         repeat split; eauto.
         eapply along_return; eauto. 
 Case "T_Elevate". 
       pose proof (IHHT c H0 env0 V).
       destruct (R_halts H1) as [v [? ?]].
       pose proof (multistep_preserves_R _ _ _ H2 H1).
       destruct H4 as [? [? ?]].
       pose proof (Congruence_ST_Elevate SM' _ _ H2).
       assert (halts (televate SM' (msubst env0 t))).
              eapply multistep_preserves_halting.
              eapply H7.
              eapply values_halt.
              constructor; eauto.
       split; [econstructor; [eapply H|eapply R_typable_empty;eauto]
              |split;[assumption|
                      intros]].
       pose proof (H6 Sto (Rsto_more_permissive _ _ Sto H H9)).
       destruct H10 as [v' [Sto' [? [? [? ?]]]]].
       pose proof (Congruence_STM_ElevateRet SM SM' v' Sto' H11).
       pose proof (Congruence_STM_Elevate SM' _ _ Sto Sto' H3 H10).
       assert (HG1:  (televate SM' v, Sto)  ~~~>*  (treturn (MNonReact SM') v', Sto')) by (eauto).
       pose proof (Congruence_LC _ _ Sto H7).
       assert (HG2: (televate SM' (msubst env0 t), Sto) ~~~>* (treturn (MNonReact SM') v', Sto')) by (eauto).
       exists v'; exists Sto'.
       repeat (split; auto).
       eapply Rsto_less_permissive; eauto.
 Case "T_Get".
       unfold R; fold Rsto; fold R.
       repeat split.
       constructor;auto.
       eapply values_halt; constructor;auto.
       intros.
       destruct H1 as [? [? ?]].
       destruct H3 as [t' [Sto' [? [? ?]]]].
       destruct (R_halts H3) as [v' [? ?]].
       pose proof (STM_Get (MStateT T E SM) Sto' t').
       pose proof (Congruence_Treturn t' v' (MNonReact (MStateT T E SM)) Sto H6).
       rewrite <- H5 in H8.
       assert (HG1:  (tget (MStateT T E SM), Sto) ~~~>* (treturn (MNonReact (MStateT T E SM)) v', Sto)) by (
              eapply rtc_R_mo in H8; eapply rtc_Trans_mo; eauto).
       exists v'; exists Sto.
       pose proof (multistep_preserves_R _ _ _ H6 H3).
       repeat split; auto.
       exists t'; exists Sto'.
       repeat (split; auto).
 Case "T_Put".
       pose proof (IHHT c H0 env0 V).
       destruct (R_halts H1) as [v [? ?]].
       unfold R; fold Rsto; fold R.
       pose proof (Congruence_ST_Put (MStateT T E SM) _ _ H2).
       repeat split.
       constructor;auto.
       eapply R_typable_empty;eauto.
       eapply multistep_preserves_halting.
              eauto.
       eapply values_halt; constructor; auto.
       intros.
       destruct H5.
       destruct H6 as [? [t' [Sto' [? [? ?]]]]].
       pose proof (STM_Put v (MStateT T E SM) Sto' t' H3).
       exists tunit; exists (v::Sto').
       rewrite <- H9 in H10.
       eapply rtc_R_mo in H10.
       pose proof (Congruence_LC _ _ Sto H4).
       pose proof (multistep_preserves_R _ _ _ H2 H1).
       pose proof (R_typable_empty H12).
       repeat split.
           eapply rtc_Trans_mo; eauto.
           constructor.
           constructor.
           eapply values_halt.
           constructor.
           constructor.
           assumption. 
           eapply Rsto_all_values;eauto. 
           constructor;auto.
           eapply Rsto_matches.
           assumption.
           exists v; exists Sto'.
           repeat split; auto.
 Case "T_RunSt".         
   pose proof (IHHT1 c H env0 V);
   pose proof (IHHT2 c H env0 V);
   destruct (R_halts H0) as [v1 [? ?]]. 
   assert (R (TMonadic (MNonReact (MStateT TS E SM)) T) v1) by (eapply multistep_preserves_R;eauto).
   destruct H0. destruct H5.
   pose proof (R_typable_empty H1).
   destruct (R_halts H1) as [v2 [? ?]].
   assert (trunst (msubst env0 t1) (msubst env0 t2) ~~>* trunst v1 v2) by (
          pose proof (Congruence_RunSt1 _ _ (msubst env0 t2) H2); 
          pose proof (Congruence_RunSt2 _ _ _ H3 H8);
          eauto).
   assert (halts (trunst (msubst env0 t1) (msubst env0 t2))) by (
          assert (halts (trunst v1 v2)) by (eapply values_halt; constructor; auto);
          eapply multistep_preserves_halting; eauto).
   pose proof (multistep_preserves_R _ _ _ H8 H1). 
   split; [econstructor; eauto | split; [ 
           assumption | intros ]].
   assert (Rsto (MStateT TS E SM) (v2::Sto)).
           unfold Rsto; fold R; fold Rsto.
           repeat split. 
           constructor; auto.
                  eapply Rsto_all_values;eauto.
           eapply matches_mo_statet.
           apply (R_typable_empty H12).
           eapply Rsto_matches; eauto.
           exists v2; exists Sto.
           repeat (split; auto).
   pose proof (H6 _ H14).
   destruct H15 as [v3 [Sto' [? [? [? ?]]]]].
   pose proof (Congruence_LC _ _ (v2 :: Sto) H2). 
   assert (done_mo ((treturn (MNonReact (MStateT TS E SM)) v3, Sto'))) by (constructor;auto).
   pose proof (step_same_done _ _ _ H20 H19 H15).
   unfold Rsto in H18; fold R in H18; fold Rsto in H18.
   destruct H18. destruct H22 as [HSto_match [t'' [Sto'' [? [? ?]]]]].
   rewrite -> H24 in H21. 
   assert (Hall_sto: store_all_values Sto) by (eapply Rsto_all_values;eauto).
   pose proof (Congruence_STM_RunSt v1 (treturn (MNonReact (MStateT TS E SM)) v3) v2 t'' Sto Sto'' H9 Hall_sto H21). 
   pose proof (Congruence_LC _ _ Sto H10).
   destruct (R_halts H22) as [v'' [? ?]].
   assert ((trunst (treturn (MNonReact (MStateT TS E SM)) v3) t'' ~~>* (trunst (treturn (MNonReact (MStateT TS E SM)) v3) v''))).
          eapply Congruence_RunSt2. constructor; auto. auto.
   pose proof (Congruence_LC _ _ Sto'' H29).
   pose proof (Congruence_STM_RunStRet v3 v'' Sto'' TS E SM H16 H28).
   assert ((trunst (msubst env0 t1) (msubst env0 t2), Sto) ~~~>*
(treturn (MNonReact SM) (tpair v3 v''), Sto'')).
         eapply rtc_Trans_mo; eauto.
         eapply rtc_Trans_mo; eauto.
         eapply rtc_Trans_mo; eauto.
   assert (R TS v'') by (eapply multistep_preserves_R;eauto).
   exists (tpair v3 v''); exists Sto''.
        repeat split; auto.
        constructor; auto.  
        apply (R_typable_empty H17).
        apply (R_typable_empty H33).
        eapply values_halt. constructor; auto.
        exists v3; exists v''.
            repeat split; eauto.
  Case "T_RunId".
    pose proof (IHHT c H env0 V).
    destruct (R_halts H0) as [v [? ?]]. 
    eapply multistep_preserves_R'.
    constructor.
    apply R_typable_empty; auto.
    eapply Congruence_RunId.
    eapply H1.
    pose proof (multistep_preserves_R _ _ _ H1 H0).
    destruct H3 as [? [? ?]].
    assert (Rsto MIdentity nil) by (repeat constructor).
    pose proof (H5 _ H6).
    destruct H7 as [t' [Sto' [? [? [? ?]]]]].
    inversion H10; subst. 
    destruct H12.
    inversion H12; subst.
    eapply multistep_preserves_R'.
    constructor; auto.
    eapply Congruence_RunIdMo; eauto.
    eapply step_preserves_R'.
    repeat econstructor; eauto.
    apply R_typable_empty; auto.
    eapply ST_RunIdRet; eauto.
    assumption.
  Case "T_Pause".
    pose proof (IHHT c H env0 V).
    destruct (R_halts H0) as [v [? ?]].
    eapply multistep_preserves_R'.
      econstructor; eapply R_typable_empty; eauto.
      eapply Congruence_ST_Pause; eauto.
      split.
    econstructor. eapply R_typable_empty. eapply multistep_preserves_R; eauto.
    split.
    apply values_halt; constructor; auto.
    intros.
    destruct H0 as [? [? ?]].
    destruct (H5 _ H3) as [v' [Sto' [? [? [? ?]]]]].
    destruct H8 as [? [? ?]].
    destruct H11 as [v1 [v2 [? [? [? [? ?]]]]]].
    destruct H15 as [? [? ?]].
    pose proof (Congruence_Treturn _ _ (MNonReact SM) Sto' H11).
    exists (tpause (MReactT TI TO SM) T v), Sto.
  split; auto.
  split.
    constructor; auto.
  split; auto.
  eapply along_pause.
    eapply rtc_refl.
    auto. eapply H12. eapply H13.
    auto.
    eapply H9.
    eapply step_same_done.
    constructor;auto.
    eapply (Congruence_LC _ _ Sto H1).
    eauto.
    auto.
    intros.
       apply (R_halts (H17 _ H19)).
    intros.       
       pose proof (H17 _ H19).
       destruct H21 as [? [? ?]].
       destruct (H23 _ H20) as [? [? [? [? [? ?]]]]].
       eapply step_preserves_along_star; eauto.
  Case "T_Unfold".
    pose proof (IHHT1 c H env0 V).
    pose proof (IHHT2 c H env0 V).
    destruct (R_halts H0) as [v1 [? ?]].
    destruct (R_halts H1) as [v2 [? ?]].
    pose proof (multistep_preserves_R _ _ _ H2 H0).
    pose proof (multistep_preserves_R _ _ _ H4 H1).
    assert (HTunfold1:  (tunfold (MReactT TI TO SM) TA TB (msubst env0 t1) (msubst env0 t2) ~~>*
             (tunfold (MReactT TI TO SM) TA TB v1 v2))).
             pose proof (Congruence_ST_Unfold1 (MReactT TI TO SM) TA TB _ _ (msubst env0 t2) H2);
             pose proof (Congruence_ST_Unfold2 (MReactT TI TO SM) TA TB _ _ _ H3 H4); eauto. 
    assert (halts (tunfold (MReactT TI TO SM) TA TB v1 v2)).
           eapply values_halt; constructor; auto.
    unfold R; fold Rsto; fold R.
    split; [constructor;[apply (R_typable_empty H0)
                        |apply (R_typable_empty H1)] 
           |split; [eapply multistep_preserves_halting; eauto
                   |intros]].
    exists (tunfold (MReactT TI TO SM) TA TB v1 v2); exists Sto.
    split; [eapply Congruence_LC;eauto
           |split; [constructor;auto
                   |split; [auto
                           |]]].
    pose proof (Congruence_LC _ _ Sto HTunfold1).
    pose proof (tunfold_along _ Sto TA TB TI TO SM v1 v2 Sto H10 H6 H7 H9). 
    eapply step_preserves_along_star';eauto.
  Case "T_RunRe".
    pose proof (IHHT c H env0 V).
    destruct (R_halts H0) as [v [? ?]]. 
    pose proof (multistep_preserves_R _ _ _ H1 H0). 
    eapply multistep_preserves_R'.
         constructor. 
         apply (R_typable_empty H0).
         eapply (Congruence_ST_RunRe T _ _ H1).
     split; [constructor; apply (R_typable_empty H3)
         |split; [eapply values_halt; constructor; auto
                 |intros]].
    unfold R in H3; fold Rsto in H3; fold R in H3.
    destruct H3 as [? [? ?]].       
    pose proof (H6 _ H4).
    destruct H7 as [t' [Sto' [? [? [? ?]]]]].
    assert ((v, Sto) |> (TMonadic (MReactT TI TO SM) T)).
            constructor.
            auto.
            eapply Rsto_all_values;eauto.
            constructor.
            eapply Rsto_matches;auto.
    inv H10.
  SCase "going to treturn". 
   assert ((treturn (MReactT TI0 TO0 SM0) t'0, Sto'0) |> (TMonadic (MReactT TI TO SM) T)).
      eapply preservation_mo_star.
      eapply H14.
      eapply preservation_mo_star.
      eapply H7.
      assumption.
    inv H12. inv H19.
    pose proof (Congruence_STM_RunRe T _ _ Sto Sto' H2 H7).
    pose proof (Congruence_STM_RunReRet t'0 TI TO SM T Sto'0 H15).
    pose proof (Congruence_STM_RunRe T _ _ Sto' Sto'0 H8 H14).
    exists (tinl (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))) t'0), Sto'0.
    split. repeat (eapply rtc_Trans_mo; eauto).
           split.
           constructor; auto.
    split; auto.
    split. constructor;auto.      
           split; [eapply values_halt; constructor; auto | ]. 
           exists t'0. 
           split; [assumption|left;split;[eapply rtc_refl|assumption]].
  SCase "going to pause".
     assert ((tpause (MReactT TI0 TO0 SM0) T0 t'0, Sto'0) |> (TMonadic (MReactT TI TO SM) T)).
      eapply preservation_mo_star.
      eapply H14.
      eapply preservation_mo_star.
      eapply H7.
      assumption.
    inv H12. inv H25. 
    pose proof (Congruence_STM_RunRe T _ _ Sto Sto' H2 H7).
    pose proof (Congruence_STM_RunRe T _ _ Sto' Sto'0 H8 H14).
    pose proof (Congruence_STM_RunRePause t'0 TI TO SM T Sto'0 H15).
        assert ((trunre T v,Sto) |> (TMonadic (MNonReact SM) (TSum T (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))))).
            constructor;auto.
            econstructor;auto. 
            eapply Rsto_all_values;eauto.
            eapply Rsto_matches. 
            auto.
    assert ((t'0 >>=[TSum T
                   (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))] 
          (\(0:TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))) 
            treturn (MNonReact SM) (tinr T (tvar 0))), Sto'0) |> (TMonadic (MNonReact SM) (TSum T (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))))).
             eapply preservation_mo_star; eauto.
             eapply preservation_mo_star; eauto.
             eapply preservation_mo_star; eauto.
    assert (value ( (\(0:TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))) 
                      treturn (MNonReact SM) (tinr T (tvar 0))))) by constructor.
    assert (value ((tpair vl vr))) by (constructor;auto).
    pose proof (Congruence_Mo_Bind1 (TSum T (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))) _ _ _ _ _ H32 H20).
    pose proof (Congruence_BindRet  (TSum T (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))) _ _ (MNonReact SM) Sto'' H33 H32).
    pose proof (Congruence_AppAbs_Mo 0 (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))
                                       (treturn (MNonReact SM) (tinr T (tvar 0))) _ Sto'' H32).
    simpl in H36.
  (* pose proof (rtc_Trans_mo _ _ _ H13
               (rtc_Trans_mo _ _ _ H22
               (rtc_Trans_mo _ _ _ H24
               (rtc_Trans_mo _ _ _ H34
               (rtc_Trans_mo _ _ _ H35 H36))))). *)
    exists (tinr T (tpair vl vr)), Sto''.
           split;auto. 
           eapply rtc_Trans_mo. 
                  eapply H13.
           eapply rtc_Trans_mo.
                  eapply H22.
           eapply rtc_Trans_mo.
                  eapply H24.
           eapply rtc_Trans_mo.
                  eapply H34.
           eapply rtc_Trans_mo.
                  eapply H35.
           eapply rtc_Trans_mo.
                  eapply Congruence_AppAbs_Mo.
                         auto.
                         simpl.
                         eapply rtc_refl.
           split; try (constructor;auto;fail).
           split; auto.
    assert ((treturn (MNonReact SM) (tpair vl vr), Sto'') |> TMonadic (MNonReact SM)
           (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
           eapply preservation_mo_star;eauto.
                  constructor. auto.
                  eapply Rsto_all_values;eauto.
                  eapply Rsto_matches;eauto.
    inv H37.
    inv H42.
    inv H40.
    split.
       constructor.
       constructor;auto.
    split.
       apply values_halt; (constructor;auto).
    eexists. split. eapply H33.
             right.
             split.
                eauto.
    split.
       constructor;auto.
    split.
       apply values_halt; (constructor;auto).
    do 2 eexists; repeat split; auto.
       apply values_halt;auto.
    econstructor; eauto.
    apply (R_typable_empty H38).
    intros.
      destruct (H26 _ H38) as [vapp []].
      pose proof (Congruence_LC _ _ Sto0 H41).
      pose proof (H27 _ _ H38 H39).
      exists vapp, Sto0.
             repeat split; auto.
             eapply step_preserves_along_star';eauto.
Qed.

Theorem normalization : forall (t:tm) (T:Ty),
        \empty |-- t \in T -> halts t.
Proof.
  intros t T H.
  replace t with (msubst nil t).
  eapply R_halts. 
  eapply msubst_R; eauto. instantiate (2:= nil). eauto.  
  eapply V_nil.
  auto.
Qed.

Lemma WT_Configs_RSto_react : forall (t:tm) (Sto:store) (TI TO T:Ty) (SM:SMo),
                     (t, Sto) |> TMonadic (MReactT TI TO SM) T -> 
                     Rsto SM Sto.
Proof.
  intros t Sto TI TO T SM.
  generalize dependent SM TI TO T t Sto.
  induction SM; intros ? ? ? ? ? Htco.
  Case "MIdentity".
    inversion Htco; subst. inversion H5; subst.
    split; auto.
  Case "MStateT".
    inversion Htco; subst. 
    inversion H5;   subst. 
    inversion H1; subst. simpl. 
    split; auto.
    split; auto.
    exists t1, Sto0.
    split; auto. 
      replace t1 with (msubst nil t1).
        eapply msubst_R. instantiate (1:=nil). simpl. assumption. apply V_nil.
        auto.
      split; auto.
      assert ((treturn (MReactT TI TO SM) t1,Sto0) |> TMonadic (MReactT TI TO SM) t).
             constructor.
             constructor.
             auto. 
             inv H4.
             constructor.
             auto.
      apply (IHSM TI TO t (treturn (MReactT TI TO SM) t1) Sto0 H).
Qed.

Lemma WT_Configs_RSto_nonreact : forall (t:tm) (Sto:store) (T : Ty) (SM:SMo),
                     (t, Sto) |> TMonadic (MNonReact SM) T -> 
                     Rsto SM Sto.
Proof.
  intros t Sto T SM.
  generalize dependent SM T t Sto.
  induction SM; intros ? ? ? Htco.
  Case "MIdentity".
    inversion Htco; subst. inversion H5; subst.
    split; auto.
  Case "MStateT".
    inversion Htco; subst.
    inversion H5; subst. 
    split; auto.
    split; auto.
    exists t1, Sto0.
    split; auto. 
      replace t1 with (msubst nil t1).
        eapply msubst_R. instantiate (1:=nil). simpl. assumption. apply V_nil.
        auto.
      split; auto.
      assert ((treturn (MNonReact SM) t1,Sto0) |> TMonadic (MNonReact SM) t).
             constructor.
             constructor.
             auto.
             inv H4.
             auto.
      apply (IHSM t (treturn (MNonReact SM) t1) Sto0 H).
Qed.

Theorem normalization_mo : forall (t:tm) (Sto:store) (T:Ty) (M:Mo),
        (t, Sto) |> TMonadic M T -> 
                         halts_mo (t, Sto).
Proof.
  intros. destruct M.
          pose proof (WT_Configs_RSto_react t Sto t0 t1 T s H).  
          replace t with (msubst nil t).
          eapply R_halts_react. 
          eapply msubst_R; eauto. 
          inversion H; subst.
          eapply typable_empty_everywhere.
                 eapply H5.
                 instantiate (1 := nil).
                 apply V_nil.
                 auto.
                 auto.
          pose proof (WT_Configs_RSto_nonreact t Sto T s H).  
          replace t with (msubst nil t).
          eapply R_halts_nonreact. 
          eapply msubst_R; eauto. 
          inversion H; subst.
          eapply typable_empty_everywhere.
                 eapply H5.
                 instantiate (1 := nil).
                 apply V_nil.
                 auto.
                 auto.
Qed.

Theorem non_reactive_done_mo : forall t Sto T SM,
                                 (t,Sto) |> TMonadic (MNonReact SM) T ->
                                 forall v Sto',
                                   done_mo (v,Sto') -> 
                                   (t,Sto) ~~~>* (v,Sto') ->
                                   exists v',
                                   v = treturn (MNonReact SM) v'.
Proof.
  intros.
  destruct (normalization_mo _ _ _ _ H) as [(v1,Sto1) [? ?]].
  preservation.
  inv H3.
  inv H4.
  inv H11.
  exists v0.
     pose proof (done_unique _ _ _ H0 H3 H1 H2).
     inv H6.
  preservation.
  inv H3.
  inv H4.
  inv H14.
Qed.

(** Simple naive tactic to normalize well-type expressions. *)

Ltac lc_normalize E :=
   let v := fresh "v"  in (
       let HS := fresh "HS" in (
           let HV := fresh "HV" in (
           match type of E with tm =>
                 match goal with 
                 | [H: (has_type \empty E ?T) |- _ ] => 
                       destruct (normalization _ _ H) as [v [HS HV]]
                 end
           end))).


Ltac mo_normalize E :=
   let v := fresh "v" in (
       let HS := fresh "HS" in (
           let HV := fresh "HV" in (
           match type of E with tm =>
                 match goal with 
                 | [H: (configuration_has_type (E,?Sto') ?T) |- _] =>  
                       let Sto := fresh "Sto" in 
                           (destruct (normalization_mo _ _ _ _ H) as [(v,Sto) [HS HV]])
                 end
           end))).
