Require Import Lib.
Require Import ReWire_Base.
Require Export Relations.Relation_Definitions.
Require Import Coq.Classes.EquivDec. 
Require Import Coq.Logic.FunctionalExtensionality.

Instance IdDec : @EqDec id eq eq_equivalence.
Proof. exact eq_id_dec. Defined. 

(** Now we define a function FVs that returns any variables  
    that occur free in a term . *)
 
Function FVs (t : tm) {struct t} : list id := 
 match t with
 | (tvar x)              => (x :: nil)
 | tapp t1 t2            => (FVs t1) ++ (FVs t2)
 | tabs x T t'           => (FVs t')/{x}
 | tunit                 => (@nil id)
 | tpair t1 t2           => (FVs t1) ++ (FVs t2)
 | tproj t1 t2           => (FVs t1) ++ (FVs t2)
 | tinl T t1             => (FVs t1)
 | tinr T t1             => (FVs t1)
 | tcase t1 t2 t3        => (FVs t1) ++ (FVs t2) ++ (FVs t3)
 | treturn M t1          => (FVs t1)
 | tbind T t1 t2         => (FVs t1) ++ (FVs t2)
 | tlift M t1            => (FVs t1)
 | televate SM t1        => (FVs t1)
 | tget SM               => (@nil id)
 | tput SM t1            => (FVs t1)
 | trunst t1 t2          => (FVs t1) ++ (FVs t2)
 | trunid t1             => (FVs t1)
 | tpause M T t1         => (FVs t1)
 | tunfold M TA TB t1 t2 => (FVs t1) ++ (FVs t2)
 | trunre T t1           => (FVs t1)
 end.

Theorem fv_dec : forall x t, {In x (FVs t)} + {~ In x (FVs t)}. 
Proof. intros. apply id_dec_in.
Defined.

Theorem eq_tm_dec : forall t1 t2 : tm, {t1 = t2} + {t1 <> t2}.
Proof. decide equality; 
  first [ apply Ty_eq_dec 
        | apply Mo_eq_dec
        | apply SMo_eq_dec 
        | apply eq_id_dec ].
Qed.

Function convert (n : id) : nat :=
  match n with
  | (Id m)     => m
  end.

Lemma convert_unique_aux : 
  forall n : id,
  exists x : nat,
     convert n = x /\
     (forall x' : nat, convert n = x' -> x = x').
Proof. intuition.
       functional induction (convert n).
       exists m.
       split;auto.
Qed.

Lemma convert_unique : forall n : id, 
      exists ! (m : nat), convert n = m. 
Proof. apply convert_unique_aux. Qed.

Definition convert_list := fun l => map (convert) l.

Definition find_new_id (l : list id) : nat := 
  let l := convert_list l in
   (1 + (fold_right (fun n x => max n x) 0 l)).

Lemma find_new_nonzero : forall l, find_new_id l <> 0.
Proof. unfold find_new_id. intros; auto with arith. 
Qed. 

Definition find_max_id (l : list id) : nat :=
  let l := convert_list l in
  fold_right (fun n x => max n x) 0 l. 

Lemma find_new_max : forall l x,
      In x l -> (convert x <= find_max_id l).
Proof. induction l; simpl; intros. 
       elim H. 
       elim H; intros. 
       subst x. unfold find_max_id.  simpl. auto with arith.
       apply le_trans with (find_max_id l); auto with arith.
                      unfold find_max_id;  simpl.
                      auto with arith.
Qed.

Lemma find_new_id_new : forall l,
      let k := (Id (find_new_id l)) in ~ In k l.
Proof. red; unfold find_new_id; intuition.
       pose proof (find_new_max _ _ H);
            unfold find_max_id in *. 
       apply le_Sn_n in H0; auto.
Qed.
             
(** * Substitution *)

Reserved Notation "'[' x ':=' s ']' t" (at level 20).

(* Substitution function. *)
Fixpoint subst (x:id) (s:tm) (t:tm) : tm :=
  match t with
  | tvar x' => 
      if eq_id_dec x x' then s else t
  | tapp t1 t2 => 
      tapp ([x:=s] t1) ([x:=s] t2)
  | tabs x' T t1 => 
      tabs x' T (if eq_id_dec x x' then t1 else ([x:=s] t1))
  | tunit =>
      tunit
  | tpair t1 t2 =>
      tpair ([x:=s] t1) ([x:=s] t2)
  | tproj t1 t2 =>
      tproj ([x:=s] t1) ([x:=s] t2)
  | tinl T t1 =>
      tinl T ([x:=s] t1)
  | tinr T t1 =>
      tinr T ([x:=s] t1)
  | tcase t1 t2 t3 =>
      tcase ([x:=s] t1) ([x:=s] t2) ([x:=s] t3)
  | treturn M t1 =>
      treturn M ([x:=s] t1)
  | tbind T t1 t2 =>
      tbind T ([x:=s] t1) ([x:=s] t2)
  | tlift M t1 =>
      tlift M ([x:=s] t1)
  | televate SM t1 =>
      televate SM ([x:=s] t1)
  | tget SM =>
      tget SM
  | tput SM t1 =>
      tput SM ([x:=s] t1)
  | trunst t1 t2 =>
      trunst ([x:=s] t1) ([x:=s] t2)
  | trunid t1 =>
      trunid ([x:=s] t1)
  | tpause M T t1 =>
      tpause M T ([x:=s] t1)
  | tunfold M TA TB t1 t2 =>
      tunfold M TA TB ([x:=s] t1) ([x:=s] t2)
  | trunre T t1 =>
      trunre T ([x:=s] t1)
  end

where "'[' x ':=' s ']' t" := (subst x s t). 

(** A variable _x_ appears free in a term _t_ .*)

Inductive appears_free_in : id -> tm -> Prop :=
  | afi_var : forall x,
      appears_free_in x (tvar x)
  | afi_app1 : forall x t1 t2,
      appears_free_in x t1 -> 
      appears_free_in x (tapp t1 t2)
  | afi_app2 : forall x t1 t2,
      appears_free_in x t2 -> 
      appears_free_in x (tapp t1 t2)
  | afi_abs : forall x y T11 t12,
      y <> x  ->
      appears_free_in x t12 ->
      appears_free_in x (tabs y T11 t12)
  (* pairs *)
  | afi_pair1 : forall x t1 t2,
      appears_free_in x t1 ->
      appears_free_in x (tpair t1 t2)
  | afi_pair2 : forall x t1 t2,
      appears_free_in x t2 ->
      appears_free_in x (tpair t1 t2)
  | afi_proj1 : forall x t1 t2,
      appears_free_in x t1 ->
      appears_free_in x (tproj t1 t2)
  | afi_proj2 : forall x t1 t2,
      appears_free_in x t2 ->
      appears_free_in x (tproj t1 t2)
  (* copairs *)
  | afi_tinl : forall x t T,
      appears_free_in x t ->
      appears_free_in x (tinl T t)
  | afi_tinr : forall x t T,
      appears_free_in x t ->
      appears_free_in x (tinr T t)
  | afi_tcase1 : forall x t1 t2 t3,
      appears_free_in x t1 ->
      appears_free_in x (tcase t1 t2 t3)
  | afi_tcase2 : forall x t1 t2 t3,
      appears_free_in x t2 ->
      appears_free_in x (tcase t1 t2 t3)
  | afi_tcase3 : forall x t1 t2 t3,
      appears_free_in x t3 ->
      appears_free_in x (tcase t1 t2 t3)
  (* monadic constructors *)
  | afi_treturn : forall x t M,
      appears_free_in x t ->
      appears_free_in x (treturn M t)
  | afi_tbind1 : forall x T t1 t2,
      appears_free_in x t1 ->
      appears_free_in x (tbind T t1 t2)
  | afi_tbind2 : forall x T t1 t2,
      appears_free_in x t2 ->
      appears_free_in x (tbind T t1 t2)
  | afi_tlift : forall x t M,
      appears_free_in x t ->
      appears_free_in x (tlift M t)
  (* state monadic constructors *)
  | afi_televate : forall x t M,
      appears_free_in x t ->
      appears_free_in x (televate M t)
  | afi_tput : forall x t M,
      appears_free_in x t ->
      appears_free_in x (tput M t)
  | afi_trunst1 : forall x t1 t2,
      appears_free_in x t1 ->
      appears_free_in x (trunst t1 t2)
  | afi_trunst2 : forall x t1 t2,
      appears_free_in x t2 ->
      appears_free_in x (trunst t1 t2)
  | afi_trunid : forall x t,
      appears_free_in x t ->
      appears_free_in x (trunid t)
  (* reactive monadic constructors *)
  | afi_tpause : forall x M T t,
      appears_free_in x t ->
      appears_free_in x (tpause M T t)
  | afi_tunfold1 : forall x M TA TB t1 t2,
      appears_free_in x t1 ->
      appears_free_in x (tunfold M TA TB t1 t2)
  | afi_tunfold2 : forall x M TA TB t1 t2,
      appears_free_in x t2 ->
      appears_free_in x (tunfold M TA TB t1 t2)
  | afi_trunre : forall x T t,
      appears_free_in x t ->
      appears_free_in x (trunre T t)
.

Hint Constructors appears_free_in.
 
Ltac afi_solve :=
  match goal with 
  | [ |- appears_free_in ?x (tvar ?x) ] => 
    eapply afi_var;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tapp ?t _)] => 
    eapply afi_app1;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tapp _ ?t)] => 
    eapply afi_app2;eauto
  | [H: ?y <> ?x, H': appears_free_in ?x ?t |- appears_free_in ?x (tabs ?y _ ?t)] => 
    eapply afi_abs;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tpair ?t _)] => 
    eapply afi_pair1;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tpair _ ?t)] => 
    eapply afi_pair2;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tproj ?t _)] =>
    eapply afi_proj1;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tproj _ ?t)] => 
    eapply afi_proj2;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tinl _ ?t)] => 
    eapply afi_tinl;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tinr _ ?t)] => 
    eapply afi_tinr;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tcase ?t _ _)] =>
    eapply afi_tcase1;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tcase _ ?t _)] =>
    eapply afi_tcase2;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tcase _ _ ?t)] =>
    eapply afi_tcase3;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (treturn _ ?t)] => 
    eapply afi_treturn;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tbind _ ?t _)] => 
    eapply afi_tbind1;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tbind _ _ ?t)] => 
    eapply afi_tbind2;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tlift _ ?t)] => 
    eapply afi_tlift;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (televate _ ?t)] => 
    eapply afi_televate;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tput _ ?t)] =>
    eapply afi_tput;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (trunst ?t _)] => 
    eapply afi_trunst1;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (trunst _ ?t)] =>
    eapply afi_trunst2;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (trunid ?t)] =>
    eapply afi_trunid;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tpause _ _ ?t)] => 
    eapply afi_tpause;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tunfold _ _ _ ?t _)] => 
    eapply afi_tunfold1;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (tunfold _ _ _ _ ?t)] => 
    eapply afi_tunfold2;eauto
  | [H: appears_free_in ?x ?t |- appears_free_in ?x (trunre _ ?t) ] =>
    eapply afi_trunre;eauto
end.

Lemma afi_dec : forall x t, appears_free_in x t \/ ~ appears_free_in x t.
Proof. intro x; induction t; eauto;
       repeat match goal with
       | [ |- (appears_free_in ?x (tvar ?i)) \/ ~ (appears_free_in ?x (tvar ?i)) ] => 
         (destruct (eq_id_dec x i); [subst;left;eauto|right;red;intros Hafi;inv Hafi])
       | [IH: _ \/ _ |- _] => destruct IH
       | _                 => idtac end;
       match goal with
       | [|- (_ ?x (tabs ?i _ ?t)) \/ ~ (_ ?x (tabs ?i _ ?t))] => 
            destruct (eq_id_dec x i);
           [try (right; intros Hafi; inv Hafi; fail)
           |try (left;constructor;auto;fail)]
       | _ => idtac end;
       match goal with
       | [H: appears_free_in _ _ |- _] => left;try afi_solve
       | _                             => right; intros Hafi; inv Hafi
       end.
Qed.

Lemma FVs_AFI_eq : forall (x:id) (t:tm), In x (FVs t) <-> appears_free_in x t.
Proof. intros x t; generalize dependent t x.
       split; [Case "->"|Case "<-"]; functional induction (FVs t); intros; simpl in *;
       match goal with
         | [H: _ \/ _ |- _] => destruct H; [subst; constructor| contradiction]
         | [H: False  |- _] => contradiction
         | [H: In ?x (FVs ?t1 ++ FVs ?t2) |- _ ] => (destruct (in_app_or _ _ x H);
                                                     match goal with
                                                       | [H: (In ?x (FVs ?t1) -> _), H': (In ?x (FVs ?t1)) |- _ ] => specialize (H H');afi_solve end)
         | [H: In ?x (FVs ?t/{?x0}) |- _] => destruct (eq_id_dec x0 x); [subst;contradict H; apply notin_rem_id|pose proof (IHl (in_rem_id _ _ _ H)); constructor; auto]
         | _ => try (constructor; exact (IHl H); fail) || (try (destruct (in_app_or _ _ x H); [specialize (IHl H0); afi_solve
                                          |destruct (in_app_or _ _ x H0);
                                            [specialize (IHl0 H1); afi_solve
                                            |specialize (IHl1 H1); afi_solve]]; fail)) || (try (inv H; fail))
       end;
       match goal with
         | [H: appears_free_in ?x _ |- In ?x (FVs ?t1 ++ FVs ?t2) ] => inv H;
             (match goal with
                | [H: (appears_free_in ?x ?t1 -> _), H': appears_free_in ?x ?t1 |-  In ?x (FVs ?t1 ++ FVs ?t2) ] => apply in_app_l; exact (H H')
                | [H: (appears_free_in ?x ?t2 -> _), H': appears_free_in ?x ?t2 |-  In ?x (FVs ?t1 ++ FVs ?t2) ] => apply in_app_r; exact (H H')
              end)
         | _ => idtac
       end.
       inv H.
       apply in_rem_neq; [firstorder|exact (IHl H5)].
       inv H.
           apply in_app_l. exact (IHl H2).
           apply in_app_r.
                 apply in_app_l. exact (IHl0 H2).
           apply in_app_r.
                 apply in_app_r. exact (IHl1 H2).
Qed.                 
           
Definition closed (t:tm) := forall x, ~ appears_free_in x t.
       
Lemma context_invariance : forall Gamma Gamma' t S,
      Gamma |-- t \in S  ->
      (forall x, appears_free_in x t -> Gamma x = Gamma' x)  ->
      Gamma' |-- t \in S.
Proof.
  intros; generalize dependent Gamma';
  induction H; intros Gamma' Heqv; auto; try (econstructor;eauto;fail); 
  [ apply T_Var; eauto; rewrite <- Heqv;eauto
  | apply T_Abs; apply IHhas_type; intros y Hafi;
    unfold extend; destruct (eq_id_dec x y); eauto].
Qed.

Lemma free_in_context : forall x t T Gamma,
      appears_free_in x t ->
        Gamma |-- t \in T ->
        (exists T', Gamma x = Some T').
Proof.
  intros x t T Gamma Hafi Htyp;
  induction Htyp; inversion Hafi; subst; eauto;
    destruct IHHtyp as [? Hctx]; auto; 
    unfold extend in Hctx;
    rewrite neq_id in Hctx; eauto.
Qed.

Lemma subst_preserves_typing : forall Gamma x U v t T,
      (extend Gamma x U) |-- t \in T ->
                  \empty |-- v \in U ->
                   Gamma |-- ([x:=v] t) \in T.
Proof.
  intros Gamma x U v t S Htypt Htypv; 
  generalize dependent Gamma; generalize dependent S;
  induction t; intros S Gamma Htypt; simpl; inversion Htypt; subst; try (econstructor;eauto;fail);
  [simpl; rename i into y; destruct (eq_id_dec x y)
  |rename i into y; rename t into T11;
   apply T_Abs; eauto; destruct (eq_id_dec x y)];
  [   subst; 
      unfold extend in H1; rewrite eq_id in H1; 
      inversion H1; subst; clear H1;
      eapply context_invariance; eauto;
      intros x Hcontra;
      destruct (free_in_context _ _ S empty Hcontra) as [T' HT']; eauto;
      inversion HT' |
      apply T_Var; eauto; unfold extend in H1; rewrite neq_id in H1; eauto |
      eapply context_invariance; eauto; subst; intros x Hafi; unfold extend;
      destruct (eq_id_dec y x); eauto |
      apply IHt;  eapply context_invariance; eauto; intros z Hafi;
      unfold extend; destruct (eq_id_dec y z);eauto; subst; rewrite neq_id; eauto].
Qed.


Corollary typable_empty_closed : forall t T, 
          \empty |-- t \in T  ->
          closed t.
Proof.
  intros. unfold closed. intros x H1.
  destruct (free_in_context _ _ _ _ H1 H) as [T' HC].
  inversion HC.  
Qed.

Lemma vacuous_substitution : forall  t x, 
     ~ appears_free_in x t  ->
     forall t', [x:=t']t = t.
Proof.
  intros t x H t' ; induction t; simpl; eauto;
   try (match goal with
   | [ |- appcontext[(eq_id_dec ?x ?i)] ] => destruct (eq_id_dec x i); eauto; try (exfalso; subst; apply H;eauto; fail)
   end);
  try (repeat (match goal with
          | [H : context[([?x := ?t']?t1) = ?t1] |- _ ] => rewrite H; eauto; hide hyp H
          end)). 
Qed.

Lemma subst_closed: forall t, 
      closed t  ->
      forall x t', [x:=t']t = t.
Proof.
  intros; apply vacuous_substitution; apply H.
Qed.

Lemma subst_not_afi : forall t x v, 
      closed v ->  
      ~ appears_free_in x ([x:=v]t).
Proof.
  unfold closed, not.
  intros t x v P Hcond; induction t;
  simpl in Hcond; try (inv Hcond; eauto; fail).
  destruct (eq_id_dec x i); [eauto|inv Hcond].
  destruct (eq_id_dec x i); eauto; [inv Hcond|inv Hcond].
Qed.
       
Lemma duplicate_subst : forall t' x t v, 
      closed v -> [x:=t]([x:=v]t') = [x:=v]t'.
Proof.
  intros; 
  eapply vacuous_substitution;
  apply subst_not_afi; eauto.
Qed.

Lemma swap_subst : forall t x x1 v v1, 
        x <> x1 -> 
       closed v ->
      closed v1 -> 
      [x1:=v1]([x:=v]t) = [x:=v]([x1:=v1]t).
Proof. 
  intros t x x1 v v1 Hneq Hclv Hclv1.
   induction t;intros; simpl; auto;
   try (repeat (match goal with
                | [H: ?X = ?Y |- _ ] => rewrite H; eauto; hide hyp H
                end); fail);
  destruct (eq_id_dec x i); destruct (eq_id_dec x1 i);
  [ subst ++; apply ex_falso_quodlibet; auto
  | subst ++; simpl; rewrite eq_id; apply subst_closed; auto
  | subst ++; simpl; rewrite eq_id; rewrite subst_closed; auto                                                         
  | simpl; rewrite neq_id; eauto; rewrite neq_id; auto
  | reflexivity|reflexivity|reflexivity|rewrite IHt; auto].
Qed.

Lemma subst_rewrite : forall (x : id) (t t' : tm) (T : Ty),
      \empty |-- t \in T ->
      ([x:=t']t = t).
Proof. 
  intros.
  apply typable_empty_closed in H.
  apply subst_closed.
  assumption.
Qed.

Lemma typable_empty_config : forall t Sto T,
      (t,Sto) |> T -> \empty |-- t \in T.
Proof.  intros; inv H. Qed.

Ltac rewrite_subst :=
     repeat (match goal with 
             | [H': appcontext[(subst ?X ?t' ?t)], H: (empty |-- ?t \in ?T) |- _ ] => 
                                    let TEMP := fresh in 
                                        (pose proof (subst_rewrite X t t' T H) as TEMP; rewrite -> TEMP in H'; clear TEMP) 
             | [H': appcontext[(subst ?X ?t' ?t)], H: ((?t,?Sto) |> ?T) |- _ ]     => 
                                    let TEMP' := fresh in
                                        let TEMP := fresh in
                                           (pose proof (typable_empty_config t Sto T H) as TEMP'); (pose proof (subst_rewrite X t t' T TEMP') as TEMP;rewrite -> TEMP in H'; clear TEMP; clear TEMP')
             | [H: (has_type empty ?t ?T) |- appcontext[(subst ?X ?t' ?t)] ]       => 
                                    let TEMP := fresh in 
                                        (pose proof (subst_rewrite X t t' T H) as TEMP; rewrite -> TEMP; clear TEMP) 
             | [H: ((?t,?Sto) |> ?T)      |- appcontext[(subst ?X ?t' ?t)] ]       => 
                                    let TEMP' := fresh in
                                        let TEMP := fresh in
                                           (pose proof (typable_empty_config t Sto T H) as TEMP'); (pose proof (subst_rewrite X t t' T TEMP') as TEMP;rewrite -> TEMP; clear TEMP; clear TEMP')
             end).


